import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from .ensemble_set_model import EnsembleSetModel
from .parameter_model import ParametersModel


class InplaceVolumesModel:

    POSSIBLE_SELECTORS = [
        "FLUID_ZONE",
        "SOURCE",
        "ENSEMBLE",
        "REAL",
        "FIPNUM",
        "SET",
        "ZONE",
        "REGION",
        "FACIES",
        "LICENSE",
        "SENSNAME_CASE",
        "SENSNAME",
        "SENSCASE",
        "SENSTYPE",
    ]

    VOLCOL_ORDER = [
        "STOIIP",
        "GIIP",
        "ASSOCIATEDOIL",
        "ASSOCIATEDGAS",
        "BULK",
        "NET",
        "PORV",
        "HCPV",
    ]

    def __init__(
        self,
        volumes_table: pd.DataFrame,
        parameter_table: Optional[pd.DataFrame] = None,
        non_net_facies: Optional[List[str]] = None,
        drop_constants: bool = True,
        volume_type: str = "static",
    ):
        self._volume_type = volume_type
        self.pmodel = ParametersModel(
            parameter_table, drop_constants=drop_constants, keep_numeric_only=False
        )
        selectors = [x for x in volumes_table.columns if x in self.POSSIBLE_SELECTORS]

        if volume_type != "dynamic":
            # compute water zone volumes if total volumes are present
            if any(col.endswith("_TOTAL") for col in volumes_table.columns):
                volumes_table = self._compute_water_zone_volumes(
                    volumes_table, selectors
                )

            # stack dataframe on fluid zone and add fluid as column instead of a column suffix
            dfs = []
            for fluid in ["OIL", "GAS", "WATER"]:
                fluid_columns = [
                    x for x in volumes_table.columns if x.endswith(f"_{fluid}")
                ]
                if not fluid_columns:
                    continue
                df = volumes_table[selectors + fluid_columns].copy()
                df.columns = df.columns.str.replace(f"_{fluid}", "")
                df["FLUID_ZONE"] = fluid.lower()
                dfs.append(df)
            self._dataframe = pd.concat(dfs)

            # Set NET volumes based on facies if non_net_facies in input
            if non_net_facies is not None and "FACIES" in self._dataframe:
                self._dataframe["NET"] = self._dataframe["BULK"]
                self._dataframe.loc[
                    self._dataframe["FACIES"].isin(non_net_facies), "NET"
                ] = 0
        else:
            self._dataframe = volumes_table
            # Workaround the FUID ZONE needs to be defined in the
            # VolumetricAnalysis plugin - this will be fixed later!
            self._dataframe["FLUID_ZONE"] = "-"

        # If sensitivity run merge sensitivity columns into the dataframe
        if self.pmodel.sensrun:
            self._dataframe = pd.merge(
                self._dataframe, self.pmodel.sens_df, on=["ENSEMBLE", "REAL"]
            )

        # set column order
        colorder = self.selectors + self.VOLCOL_ORDER
        self._dataframe = self._dataframe[
            [x for x in colorder if x in self._dataframe]
            + [x for x in self._dataframe if x not in colorder]
        ]

        self._dataframe.sort_values(by=["ENSEMBLE", "REAL"], inplace=True)

        # ensure ensemble column consists of strings
        self._dataframe["ENSEMBLE"] = self._dataframe["ENSEMBLE"].astype(str)

        # compute and set property columns
        self._set_initial_property_columns()
        self._dataframe = self.compute_property_columns(self._dataframe)

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    @property
    def parameter_df(self) -> pd.DataFrame:
        return self.pmodel.dataframe

    @property
    def sensrun(self) -> bool:
        return self.pmodel.sensrun

    @property
    def volume_type(self) -> str:
        return self._volume_type

    @property
    def sensitivities(self) -> List[str]:
        return self.pmodel.sensitivities

    @property
    def sources(self) -> List[str]:
        return sorted(list(self._dataframe["SOURCE"].unique()))

    @property
    def realizations(self) -> List[int]:
        return sorted(list(self._dataframe["REAL"].unique()))

    @property
    def ensembles(self) -> List[str]:
        return list(self._dataframe["ENSEMBLE"].unique())

    @property
    def ensemble_sensitivities(self) -> Dict[str, list]:
        return {
            ens: (
                list(
                    self._dataframe.loc[
                        self._dataframe["ENSEMBLE"] == ens, "SENSNAME_CASE"
                    ].unique()
                )
                if "SENSNAME_CASE" in self._dataframe
                else []
            )
            for ens in self.ensembles
        }

    @property
    def property_columns(self) -> List[str]:
        return self._property_columns

    @property
    def volume_columns(self) -> List[str]:
        return [
            x
            for x in self._dataframe
            if x not in self.selectors
            and x not in self.property_columns
            and is_numeric_dtype(self._dataframe[x])
        ]

    @property
    def selectors(self) -> List[str]:
        return [x for x in self.POSSIBLE_SELECTORS if x in self._dataframe]

    @property
    def region_selectors(self) -> List[str]:
        return [
            x
            for x in ["FIPNUM", "ZONE", "REGION", "SET", "LICENSE"]
            if x in self.selectors
        ]

    @property
    def responses(self) -> List[str]:
        return self.volume_columns + self.property_columns

    @property
    def hc_responses(self) -> List[str]:
        return [
            x
            for x in ["STOIIP", "GIIP", "ASSOCIATEDGAS", "ASSOCIATEDOIL"]
            if x in self.volume_columns
        ]

    @property
    def parameters(self) -> List[str]:
        return self.pmodel.parameters

    @staticmethod
    def _compute_water_zone_volumes(
        voldf: pd.DataFrame, selectors: list
    ) -> pd.DataFrame:
        """Compute water zone volumes by subtracting HC-zone volumes from
        TOTAL volumes"""
        supported_columns = ["BULK_TOTAL", "NET_TOTAL", "PORE_TOTAL", "PORV_TOTAL"]
        # Format check
        for src, df in voldf.groupby("SOURCE"):
            volcols = [col for col in df if col not in selectors]
            if not any(col in volcols for col in supported_columns):
                continue
            if df[volcols].isnull().values.any():
                warnings.warn(
                    f"WARNING: Cannot calculate water zone volumes for source {src}, "
                    "due to wrong format in input volume file. \nTo ensure correct format "
                    "use: https://equinor.github.io/fmu-tools/fmu.tools.rms.html#fmu.tools."
                    "rms.volumetrics.merge_rms_volumetrics"
                )
                return voldf

        for col in [x.replace("_TOTAL", "") for x in voldf if x in supported_columns]:
            voldf[f"{col}_WATER"] = (
                voldf[f"{col}_TOTAL"]
                - voldf.get(f"{col}_OIL", 0)
                - voldf.get(f"{col}_GAS", 0)
            )
        return voldf

    def _set_initial_property_columns(self) -> None:
        """Create list of properties that can be computed based on
        available volume columns"""
        self._property_columns = []

        if all(
            col in self._dataframe for col in ["NET", "BULK"]
        ) and not self._dataframe["NET"].equals(self._dataframe["BULK"]):
            self._property_columns.append("NTG")
        if all(col in self._dataframe for col in ["BULK", "PORV"]):
            self._property_columns.append("PORO")
        if all(
            col in self._dataframe for col in ["NET", "PORV"]
        ) and not self._dataframe["NET"].equals(self._dataframe["BULK"]):
            self._property_columns.append("PORO_NET")
        if all(col in self._dataframe for col in ["HCPV", "PORV"]):
            self._property_columns.append("SW")

        for vol_column in ["STOIIP", "GIIP"]:
            if all(col in self._dataframe for col in ["HCPV", vol_column]):
                pvt = "BO" if vol_column == "STOIIP" else "BG"
                self._property_columns.append(pvt)

    def compute_property_columns(
        self, dframe: pd.DataFrame, properties: Optional[list] = None
    ) -> pd.DataFrame:
        """Compute property columns. As default all property columns are computed,
        but which properties to compute can be given as input"""
        dframe = dframe.copy()
        properties = self.property_columns if properties is None else properties

        if "NTG" in properties:
            dframe["NTG"] = dframe["NET"] / dframe["BULK"]
        if "PORO" in properties:
            dframe["PORO"] = dframe["PORV"] / dframe["BULK"]
        if "PORO_NET" in properties:
            dframe["PORO_NET"] = dframe["PORV"] / dframe["NET"]
        if "SW" in properties:
            dframe["SW"] = 1 - (dframe["HCPV"] / dframe["PORV"])
        if "BO" in properties:
            dframe["BO"] = dframe["HCPV"] / dframe["STOIIP"]
        if "BG" in properties:
            dframe["BG"] = dframe["HCPV"] / dframe["GIIP"]
        # nan is handled by plotly but not inf
        dframe.replace(np.inf, np.nan, inplace=True)
        return dframe

    def get_df(
        self,
        filters: Optional[Dict[str, list]] = None,
        groups: Optional[list] = None,
        parameters: Optional[list] = None,
        properties: Optional[list] = None,
    ) -> pd.DataFrame:
        """Function to retrieve a dataframe with volumetrics and properties. Parameters
        can be added to the dataframe if parameters are available in the instance.
        Filters are supported on dictionary form with 'column_name': [list ov values to keep].
        The final dataframe can be grouped by giving in a list of columns to group on.
        """
        dframe = self.dataframe.copy()

        groups = groups if groups is not None else []
        filters = filters if filters is not None else {}
        parameters = parameters if parameters is not None else []

        if parameters and self.parameters:
            columns = parameters + ["REAL", "ENSEMBLE"]
            dframe = pd.merge(
                dframe, self.parameter_df[columns], on=["REAL", "ENSEMBLE"]
            )
        if filters:
            dframe = filter_df(dframe, filters)

        prevent_sum_over = ["REAL", "ENSEMBLE", "SOURCE"]
        if groups:
            sum_over_groups = groups + [x for x in prevent_sum_over if x not in groups]

            # Need to sum volume columns and take the average of parameter columns
            aggregations = {x: "sum" for x in self.volume_columns}
            aggregations.update({x: "mean" for x in parameters})

            dframe = dframe.groupby(sum_over_groups).agg(aggregations).reset_index()
            dframe = dframe.groupby(groups).mean(numeric_only=True).reset_index()

        dframe = self.compute_property_columns(dframe, properties)
        if "FLUID_ZONE" not in groups:
            if not filters.get("FLUID_ZONE") == ["oil"]:
                dframe["BO"] = np.nan
            if not filters.get("FLUID_ZONE") == ["gas"]:
                dframe["BG"] = np.nan
        return dframe


def filter_df(dframe: pd.DataFrame, filters: dict) -> pd.DataFrame:
    """
    Filter dataframe using dictionary with form
    'column_name': [list of values to keep]
    """
    for filt, values in filters.items():
        dframe = dframe.loc[dframe[filt].isin(values)]
    return dframe


def extract_volumes(
    ensemble_set_model: EnsembleSetModel,
    volfolder: str,
    volfiles: Dict[str, Any],
) -> pd.DataFrame:
    """Aggregates volumetric files from an FMU ensemble.
    Files must be stored on standardized csv format.
    """
    dfs = []
    for volname, files in volfiles.items():
        if isinstance(files, list):
            volframes = [
                ensemble_set_model.load_csv(Path(volfolder) / volfile)
                for volfile in files
            ]
            df = merge_csv_files(volframes)
        elif isinstance(files, str):
            df = ensemble_set_model.load_csv(Path(volfolder) / files)
        else:
            raise ValueError("Wrong format of volfile value argument!")
        df["SOURCE"] = volname
        dfs.append(df)

    if not dfs:
        raise ValueError(
            f"Error when aggregating inplace volumetric files: {list(volfiles)}. "
            f"Ensure that the files are present in relative folder {volfolder}"
        )
    return pd.concat(dfs)


def merge_csv_files(volframes: List[pd.DataFrame]) -> pd.DataFrame:
    """Merge csv files on common columns"""
    common_columns = list(
        set.intersection(*[set(frame.columns) for frame in volframes])
    )
    merged_dframe = pd.DataFrame(columns=common_columns)
    for frame in volframes:
        merged_dframe = pd.merge(merged_dframe, frame, on=common_columns, how="outer")
    return merged_dframe



from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import webviz_core_components as wcc
from dash import Dash, Input, Output, dash_table, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


class AssistedHistoryMatchingAnalysis(WebvizPluginABC):
    """Visualize parameter distribution change prior to posterior \
    per observation group in an assisted history matching process.
    This is done by using a \
    [KS (Kolmogorov Smirnov) test](https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test) \
    matrix, and scatter plot/map for any given pair of parameter/observation. \
    KS values are between 0 and 1. \
    The closer to zero the KS value is, the smaller the change in parameter distribution \
    between prior/posterior and vice-versa. \
    The top 10 biggest parameters change are also shown in a table.

    ---

    * **`input_dir`:** Path to the directory where the `csv` files created \
        by the `AHM_ANALYSIS` ERT postprocess workflow are stored
    * **`ks_filter`:** optional argument to filter output to the data table based on ks value, \
        only values above entered value will be displayed in the data table. \
        This can be used if needed to speed-up vizualization of cases with \
        high number of parameters and/or observations group. Default value is 0.0.

    ---


    ?> The input_dir  \
    is where the results (csv files) from \
    the ERT `AHM_ANALYSIS` worflow are stored.
    ?> The ks_filter value should typically be between 0 and 0.5.

    """

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        input_dir: Path,
        ks_filter: float = 0.0,
    ):

        super().__init__()

        self.input_dir = input_dir
        self.theme = webviz_settings.theme
        self.ks_filter = ks_filter

        self.set_callbacks(app)

    @property
    def layout(self):
        return html.Div(
            id=self.uuid("main-div"),
            children=[
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            style={"flex": 1, "style": "45vh"},
                            children=[
                                wcc.Selectors(
                                    label="Filters",
                                    children=[
                                        wcc.Label(
                                            children="Filter observations:",
                                            style={"display": "block"},
                                        ),
                                        dcc.Input(
                                            id=self.uuid("filter1_id"),
                                            value="",
                                            type="text",
                                            debounce=True,
                                        ),
                                        wcc.Label(
                                            children="Filter parameters:",
                                            style={"display": "block"},
                                        ),
                                        dcc.Input(
                                            id=self.uuid("filter2_id"),
                                            value="",
                                            type="text",
                                            debounce=True,
                                        ),
                                    ],
                                ),
                                wcc.Selectors(
                                    label="Selected output:",
                                    children=wcc.RadioItems(
                                        id=self.uuid("choice_id"),
                                        options=[
                                            {
                                                "label": "One by one observation",
                                                "value": "ONE",
                                            },
                                            {
                                                "label": "All minus one observation",
                                                "value": "ALL",
                                            },
                                        ],
                                        value="ONE",
                                    ),
                                ),
                                wcc.Selectors(
                                    label="Parameter distribution:",
                                    children=dcc.Checklist(
                                        id=self.uuid("choice_hist_id"),
                                        options=[
                                            {
                                                "label": "Transformed dist. (if available)",
                                                "value": "TRANS",
                                            },
                                        ],
                                        value=[],
                                    ),
                                ),
                            ],
                        ),
                        wcc.Frame(
                            color="white",
                            highlight=False,
                            style={"flex": 3},
                            children=[
                                html.Div(
                                    id=self.uuid("output_graph"),
                                ),
                            ],
                        ),
                        wcc.Frame(
                            color="white",
                            highlight=False,
                            style={"flex": 3},
                            children=[
                                html.Div(
                                    id=self.uuid("click_data"),
                                ),
                            ],
                        ),
                    ]
                ),
                html.Div(
                    children=[
                        html.H4(
                            children="Table of ten highest parameters change/update Ks",
                        ),
                        html.Div(
                            id=self.uuid("generate_table"),
                        ),
                    ]
                ),
            ],
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.uuid("output_graph"), component_property="children"),
            Input(self.uuid("filter1_id"), component_property="value"),
            Input(self.uuid("filter2_id"), component_property="value"),
            Input(self.uuid("choice_id"), component_property="value"),
        )
        def _update_graph(input_filter_obs, input_filter_param, choiceplot):
            """Renders KS matrix (how much a parameter is changed from prior to posterior"""
            active_info = read_csv(
                get_path(self.input_dir / "active_obs_info.csv"), index_col=0
            )
            joint_ks = read_csv(
                get_path(self.input_dir / "ks.csv"), index_col=0
            ).replace(np.nan, 0.0)
            input_filter_obs = _set_inputfilter(input_filter_obs)
            input_filter_param = _set_inputfilter(input_filter_param)

            listtoplot = _get_listtoplot(joint_ks, choiceplot)
            joint_ks_sorted = joint_ks.filter(items=listtoplot).sort_index(axis=1)
            xx_data = list(
                joint_ks_sorted.filter(like=input_filter_obs, axis=1).columns
            )
            yy_data = list(
                joint_ks_sorted.filter(like=input_filter_param, axis=0).index
            )
            zz_data = _get_zzdata(joint_ks_sorted, yy_data, xx_data, active_info)

            if not xx_data or not yy_data:
                raise PreventUpdate

            yall_obs_data = list(
                joint_ks_sorted.filter(like=input_filter_param, axis=0).index
            )
            zall_obs_data = joint_ks.loc[yall_obs_data, ["All_obs"]].to_numpy()

            return wcc.Graph(
                id=self.uuid("heatmap_id"),
                figure={
                    "data": [
                        go.Heatmap(
                            x=xx_data,
                            y=yy_data,
                            z=zz_data,
                            type="heatmap",
                            colorscale="YlGnBu",
                            zmin=0,
                            zmax=1,
                            hoverinfo="text",
                            text=_hovertext_list(
                                xx_data, yy_data, zz_data, active_info
                            ),
                        ),
                        go.Heatmap(
                            x=["All_obs"],
                            y=yall_obs_data,
                            z=zall_obs_data,
                            type="heatmap",
                            colorscale="YlGnBu",
                            zmin=0,
                            zmax=1,
                            hoverinfo="text",
                            text=_hovertext_list(
                                ["All_obs"], yall_obs_data, zall_obs_data, active_info
                            ),
                            xaxis="x2",
                        ),
                    ],
                    "layout": self.theme.create_themed_layout(
                        {
                            "title": "KS Matrix (degree of parameter change prior to posterior)",
                            "xaxis": {
                                "title": "Observations",
                                "ticks": "",
                                "domain": [0.0, 0.9],
                                "showticklabels": True,
                                "tickangle": 30,
                                "automargin": True,
                            },
                            "yaxis": {
                                "title": "Parameters",
                                "ticks": "",
                                "showticklabels": True,
                                "tickangle": -30,
                                "automargin": True,
                            },
                            "xaxis2": {"ticks": "", "domain": [0.95, 1.0]},
                            "plot_bgcolor": "grey",
                        }
                    ),
                },
                style={"height": "45vh"},
                clickData={"points": [{"x": xx_data[0], "y": yy_data[0]}]},
            )

        @app.callback(
            Output(self.uuid("click_data"), component_property="children"),
            Input(self.uuid("heatmap_id"), component_property="clickData"),
            Input(self.uuid("choice_hist_id"), component_property="value"),
            prevent_initial_call=True,
        )
        def _display_click_data(celldata, hist_display):
            """render a histogram of parameters distribution prior/posterior or
            an average delta map prior-posterior."""
            obs = celldata["points"][0]["x"]
            param = celldata["points"][0]["y"]
            active_info = read_csv(
                get_path(self.input_dir / "active_obs_info.csv"), index_col=0
            )
            if "FIELD" in param:
                fieldparam = param.replace("FIELD_", "")
                mygrid_ok_short = read_csv(
                    get_path(
                        Path(str(self.input_dir).replace("scalar_", "field_"))
                        / f"delta_field{fieldparam}.csv"
                    )
                )
                maxinput = mygrid_ok_short.filter(like="Mean_").max(axis=1)
                deltadata = "Mean_D_" + obs
                return wcc.Graph(
                    id="2Dmap_avgdelta",
                    figure=px.scatter(
                        mygrid_ok_short,
                        x="X_UTME",
                        y="Y_UTMN",
                        color=deltadata,
                        range_color=[0, maxinput.max()],
                        color_continuous_scale="Rainbow",
                        opacity=0.9,
                        title=f"Mean_delta_posterior-prior {obs}, {param}",
                        hover_data=[
                            "X_UTME",
                            "Y_UTMN",
                            "Z_TVDSS",
                            "IX",
                            "JY",
                            deltadata,
                        ],
                    ),
                )
            post_df = read_csv(get_path(self.input_dir / f"{obs}.csv"))
            prior_df = read_csv(get_path(self.input_dir / "prior.csv"))
            if "TRANS" in hist_display:
                paraml = [ele for ele in prior_df.keys() if f"_{param}" in ele]
                if paraml != []:
                    param = paraml[0]
            fig = go.Figure()
            fig.add_trace(go.Histogram(x=prior_df[param], name="prior", nbinsx=10))
            fig.add_trace(go.Histogram(x=post_df[param], name="update", nbinsx=10))
            fig.update_layout(
                self.theme.create_themed_layout(
                    {
                        "title": (
                            "Parameter distribution for observation "
                            f"{obs} ({active_info.at['ratio', obs]})"
                        ),
                        "bargap": 0.2,
                        "bargroupgap": 0.1,
                        "xaxis": {"title": param},
                    }
                )
            )

            return wcc.Graph(id="lineplots", style={"height": "45vh"}, figure=fig)

        @app.callback(
            Output(self.uuid("generate_table"), component_property="children"),
            Input(self.uuid("choice_id"), component_property="value"),
        )
        def _generatetable(choiceplot, max_rows=10):
            """Generate output table of data in KS matrix plot"""
            misfit_info = read_csv(
                get_path(self.input_dir / "misfit_obs_info.csv"), index_col=0
            )
            list_ok = list(misfit_info.filter(like="All_obs", axis=1).columns)
            listtoplot = [ele for ele in misfit_info.columns if ele not in list_ok]
            if choiceplot == "ALL":
                listtoplot = list_ok
            active_info = read_csv(
                get_path(self.input_dir / "active_obs_info.csv"),
                index_col=0,
            )

            joint_ks = read_csv(
                get_path(self.input_dir / "ks.csv"),
                index_col=0,
            ).replace(np.nan, 0.0)
            ks_filtered = _get_ks_filtered(
                listtoplot, active_info, misfit_info, joint_ks, self.ks_filter
            )

            ks_filtered = ks_filtered.sort_values(by="Ks_value", ascending=False)

            return dash_table.DataTable(
                columns=[{"name": i, "id": i} for i in ks_filtered.columns],
                editable=True,
                style_data_conditional=[
                    {
                        "if": {
                            "filter_query": "{Active Obs}=0",
                            "column_id": "Active Obs",
                        },
                        "backgroundColor": "grey",
                        "color": "white",
                    },
                ],
                data=ks_filtered.to_dict("records"),
                sort_action="native",
                filter_action="native",
                page_action="native",
                page_current=0,
                page_size=max_rows,
            )

    @property
    def tour_steps(self) -> List[Dict[str, str]]:
        return [
            {
                "id": self.uuid("main-div"),
                "content": (
                    "This dashboard helps to analyze the update step performed "
                    "during assisted history match using the ensemble smoother method. "
                    "This can give insight into which parameters are updated due to a "
                    "specific observation type, and also which observations are causing an "
                    "update in a specific parameter. KS refers to "
                    "Kolmogorov–Smirnov."
                ),
            },
            {
                "id": self.uuid("filter1_id"),
                "content": (
                    "By entering text in this field, you will be able to filter the "
                    "observations (e.g. entering 'WBP' will include only observations "
                    "containing that string)."
                ),
            },
            {
                "id": self.uuid("filter2_id"),
                "content": (
                    "Give option to filter on parameters, similarly to how it is done "
                    "for observations."
                ),
            },
            {
                "id": self.uuid("choice_hist_id"),
                "content": (
                    "Give option for plotting parameter prior/posterior distribution. "
                    "Some parameters may have transformed equivalent like LOG10"
                ),
            },
            {
                "id": self.uuid("output_graph"),
                "content": "Renders KS matrix value between 0 and 1",
            },
            {
                "id": self.uuid("click_data"),
                "content": (
                    "Render a histogram of parameters distribution prior/posterior "
                    "or an average delta map prior-posterior"
                ),
            },
            {
                "id": self.uuid("generate_table"),
                "content": "Generate output table of data in KS matrix plot",
            },
        ]

    def add_webvizstore(self):
        scalar_csv_files = list(self.input_dir.glob("*"))
        field_csv_files = list(
            Path(str(self.input_dir).replace("scalar_", "field_")).glob("*")
        )
        return [
            (
                get_path,
                [{"path": path} for path in scalar_csv_files + field_csv_files],
            )
        ]


def _hovertext_list(xx_data, yy_data, zz_data, active_info):
    """Define hovertext info"""
    hovertext = []
    for parami, paramy in enumerate(yy_data):
        hovertext.append([])
        for obsi, obsx in enumerate(xx_data):
            hovertext[-1].append(
                f"Obs ({active_info.at['ratio', obsx]}): "
                f"{obsx}<br>Param: {paramy}<br>Ks: {zz_data[parami][obsi]}"
            )
    return hovertext


def _get_ks_filtered(listtoplot, active_info, misfit_info, joint_ks, filter_value):
    """Generate filtered KS dataframe"""
    ks_filtered = pd.DataFrame()
    for obs in listtoplot:
        tmp_ks_filter = pd.DataFrame()
        tmp_ks_filter["Ks_value"] = joint_ks[joint_ks[obs] >= filter_value][obs]
        tmp_ks_filter["Observation"] = obs
        tmp_ks_filter["Parameter"] = list(joint_ks[joint_ks[obs] >= filter_value].index)
        tmp_ks_filter["Active Obs"] = active_info[obs].to_numpy()[0]
        tmp_ks_filter["Avg Obs misfit"] = misfit_info[obs].to_numpy()[0]
        tmp_ks_filter.reset_index(
            level=None, drop=True, inplace=True, col_level=0, col_fill=""
        )
        ks_filtered = pd.concat([ks_filtered, tmp_ks_filter], ignore_index=True)
    return ks_filtered


def _get_listtoplot(joint_ks, choiceplot):
    """Generate correct observations to plot based on choice made"""
    list_ok = list(joint_ks.filter(like="All_obs", axis=1).columns)
    listtoplot = [ele for ele in joint_ks.columns if ele not in list_ok]
    if choiceplot == "ALL":
        list_ok.remove("All_obs")
        listtoplot = list_ok
    return listtoplot


def _get_zzdata(joint_ks_sorted, yy_data, xx_data, active_info):
    """Generate input values to heatmap,
    shows as missing data when 0active observations
    """
    zz_data = joint_ks_sorted.loc[yy_data, xx_data].to_numpy()
    for yid in range(len(yy_data)):
        for xid, xxd in enumerate(xx_data):
            active_obs_info = active_info.at["ratio", xxd].split(" ")
            if active_obs_info[0] == "0":
                zz_data[yid][xid] = None
    return zz_data


def _set_inputfilter(input_filter: str) -> str:
    """Set the input filter to show all data if empty"""
    return "_" if input_filter == "" else input_filter


@webvizstore
def get_path(path) -> Path:
    return Path(path)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def read_csv(path: Path, index_col: Optional[int] = None) -> pd.DataFrame:
    return pd.read_csv(path, index_col=index_col)



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class Selections(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLES = "ensembles"

    def __init__(self, ensembles: List[str]) -> None:
        super().__init__("Selections")
        self._ensembles = ensembles

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensembles",
                id=self.register_component_unique_id(self.Ids.ENSEMBLES),
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                value=[self._ensembles[0] if self._ensembles else None],
                clearable=False,
                multi=True,
            ),
        ]



import io
import json
from pathlib import Path

import pandas as pd
from webviz_config.webviz_store import webvizstore


@webvizstore
def get_path(path: str) -> Path:
    return Path(path)


@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file)


@webvizstore
def find_files(folder: Path, suffix: str) -> io.BytesIO:
    return io.BytesIO(
        json.dumps(
            sorted([str(filename) for filename in folder.glob(f"*{suffix}")])
        ).encode()
    )



# pylint: disable=no-name-in-module
from webviz_config.plugins import SurfaceWithGridCrossSection


def test_surface_with_grid_crosssection(
    dash_duo, app, shared_settings, testdata_folder
) -> None:
    plugin = SurfaceWithGridCrossSection(
        app,
        shared_settings["HM_SETTINGS"],
        gridfile=(
            testdata_folder
            / "01_drogon_ahm"
            / "realization-0"
            / "iter-0"
            / "share"
            / "results"
            / "grids"
            / "geogrid.roff"
        ),
        gridparameterfiles=[
            testdata_folder
            / "01_drogon_ahm"
            / "realization-0"
            / "iter-0"
            / "share"
            / "results"
            / "grids"
            / "geogrid--phit.roff"
        ],
        surfacefiles=[
            testdata_folder
            / "01_drogon_ahm"
            / "realization-0"
            / "iter-0"
            / "share"
            / "results"
            / "maps"
            / "topvolon--ds_extract_geogrid.gri"
        ],
        surfacenames=["Top Volon"],
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from typing import Dict, List, Optional

from webviz_subsurface_components import VectorDefinition
from webviz_subsurface_components.py_expression_eval import ParserError

from webviz_subsurface._abbreviations.reservoir_simulation import (
    simulation_unit_reformat,
    simulation_vector_description,
)
from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)
from webviz_subsurface._utils.vector_calculator import (
    ExpressionInfo,
    VectorCalculator,
    get_expression_from_name,
)

from .from_timeseries_cumulatives import (
    get_cumulative_vector_name,
    is_per_interval_or_per_day_vector,
)


def create_vector_plot_titles_from_provider_set(
    vector_names: List[str],
    expressions: List[ExpressionInfo],
    provider_set: EnsembleSummaryProviderSet,
    user_defined_vector_definitions: Dict[str, VectorDefinition],
    resampling_frequency: Optional[Frequency] = None,
) -> Dict[str, str]:
    """Create plot titles for vectors

    Create plot titles for vectors by use of provider set metadata and list of
    calculation expressions

    `Return:`
    * Dictionary with vector names as keys and the corresponding title as value
    """
    vector_title_dict: Dict[str, str] = {}

    all_vector_names = provider_set.all_vector_names()
    for vector_name in vector_names:

        # Provider vector
        if vector_name in all_vector_names:
            metadata = provider_set.vector_metadata(vector_name)
            title = simulation_vector_description(
                vector_name, user_defined_vector_definitions
            )
            if metadata and metadata.unit:
                title += f" [{simulation_unit_reformat(metadata.unit)}]"
            vector_title_dict[vector_name] = title

        # Per Interval or Per Day vector
        elif is_per_interval_or_per_day_vector(vector_name):
            title = simulation_vector_description(
                vector_name, user_defined_vector_definitions
            )

            cumulative_vector = get_cumulative_vector_name(vector_name)
            metadata = provider_set.vector_metadata(cumulative_vector)
            if resampling_frequency:
                title = f"{str(resampling_frequency.value).capitalize()} " + title
            if vector_name.startswith("PER_DAY_"):
                if metadata and metadata.unit:
                    _unit = metadata.unit + "/DAY"
                    title += f" [{simulation_unit_reformat(_unit)}]"
            if vector_name.startswith("PER_INTVL_"):
                if metadata and metadata.unit:
                    title += f" [{simulation_unit_reformat(metadata.unit)}]"
            vector_title_dict[vector_name] = title

        # Calculated vector
        else:
            expression = get_expression_from_name(vector_name, expressions)
            if expression:
                unit = create_calculated_unit_from_provider_set(
                    expression, provider_set
                )
                if unit:
                    # TODO: Expression description instead of vector name in title?
                    vector_title_dict[vector_name] = f"{vector_name} [{unit}]"
                else:
                    vector_title_dict[vector_name] = vector_name
            else:
                vector_title_dict[vector_name] = vector_name
    return vector_title_dict


def create_calculated_unit_from_provider_set(
    expression: ExpressionInfo, provider_set: EnsembleSummaryProviderSet
) -> Optional[str]:
    try:
        # Parse only for validation
        VectorCalculator.parser.parse(expression["expression"])
        unit_expr: str = expression["expression"]
        for elm in expression["variableVectorMap"]:
            metadata = provider_set.vector_metadata(elm["vectorName"][0])
            if metadata and metadata.unit:
                unit_expr = unit_expr.replace(elm["variableName"], metadata.unit)

        return unit_expr
    except ParserError:
        return None



from typing import Dict, List


def _get_well_names_combined(
    well_collections: Dict[str, List[str]],
    selected_collection_names: list,
    selected_wells: list,
    combine_type: str = "intersection",
) -> List[str]:
    """Return union or intersection of well list and well collection lists."""

    selected_collection_wells = []
    for collection_name in selected_collection_names:
        selected_collection_wells.extend(well_collections[collection_name])
    selected_collection_wells = list(set(selected_collection_wells))
    if combine_type == "intersection":
        # find intersection of selector wells and selector well collections
        well_names_combined = [
            well for well in selected_wells if well in selected_collection_wells
        ]
    else:
        # find union of selector wells and selector well collections
        well_names_combined = list(selected_collection_wells)
        well_names_combined.extend(selected_wells)
        well_names_combined = list(set(well_names_combined))

    return well_names_combined



from typing import Dict, List, Optional, Tuple

import webviz_core_components as wcc
from dash import Input, Output, State, callback, no_update
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._providers.ensemble_grid_provider import EnsembleGridProvider
from webviz_subsurface.plugins._grid_viewer_fmu._types import PROPERTYTYPE


def list_to_options(values: List) -> List:
    return [{"value": val, "label": val} for val in values]


class DataSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ID = "data-selectors"
        REALIZATIONS = "realizations"
        STATIC_DYNAMIC = "static-dynamic"
        PROPERTIES = "properties"
        DATES = "dates"
        WELLS = "wells"

    def __init__(self, grid_provider: EnsembleGridProvider) -> None:
        super().__init__("Data Selection")
        self.grid_provider = grid_provider
        self.static_dynamic_options = []
        self.static_dynamic_value = None

        if grid_provider.static_property_names():
            self.static_dynamic_options.append(
                {"label": PROPERTYTYPE.STATIC, "value": PROPERTYTYPE.STATIC}
            )
            self.static_dynamic_value = PROPERTYTYPE.STATIC

        if grid_provider.dynamic_property_names():
            self.static_dynamic_options.append(
                {"label": PROPERTYTYPE.DYNAMIC, "value": PROPERTYTYPE.DYNAMIC}
            )
            if self.static_dynamic_value is None:
                self.static_dynamic_value = PROPERTYTYPE.DYNAMIC

    def layout(self) -> List[Component]:

        return [
            wcc.SelectWithLabel(
                label="Realizations",
                id=self.register_component_unique_id(DataSettings.Ids.REALIZATIONS),
                multi=False,
                options=list_to_options(self.grid_provider.realizations()),
                value=[self.grid_provider.realizations()[0]],
            ),
            wcc.RadioItems(
                label="Grid property",
                vertical=False,
                id=self.register_component_unique_id(DataSettings.Ids.STATIC_DYNAMIC),
                options=self.static_dynamic_options,
                value=self.static_dynamic_value,
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(DataSettings.Ids.PROPERTIES),
                multi=False,
            ),
            wcc.SelectWithLabel(
                label="Timestep",
                id=self.register_component_unique_id(DataSettings.Ids.DATES),
                multi=False,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(DataSettings.Ids.PROPERTIES).to_string(),
                "options",
            ),
            Output(
                self.component_unique_id(DataSettings.Ids.PROPERTIES).to_string(),
                "value",
            ),
            Input(
                self.component_unique_id(DataSettings.Ids.STATIC_DYNAMIC).to_string(),
                "value",
            ),
        )
        def _populate_properties(
            static_dynamic: str,
        ) -> Tuple[List[Dict[str, str]], List[str]]:
            if PROPERTYTYPE(static_dynamic) == PROPERTYTYPE.STATIC:
                prop_names = self.grid_provider.static_property_names()

            else:
                prop_names = self.grid_provider.dynamic_property_names()

            return (
                [{"label": prop, "value": prop} for prop in prop_names],
                [prop_names[0]],
            )

        @callback(
            Output(
                self.component_unique_id(DataSettings.Ids.DATES).to_string(),
                "options",
            ),
            Output(
                self.component_unique_id(DataSettings.Ids.DATES).to_string(),
                "value",
            ),
            Input(
                self.component_unique_id(DataSettings.Ids.PROPERTIES).to_string(),
                "value",
            ),
            State(
                self.component_unique_id(DataSettings.Ids.STATIC_DYNAMIC).to_string(),
                "value",
            ),
            State(
                self.component_unique_id(DataSettings.Ids.DATES).to_string(),
                "options",
            ),
        )
        def _populate_dates(
            property_names: List[str],
            static_dynamic: str,
            current_date_options: List,
        ) -> Tuple[List[Dict[str, str]], Optional[List[str]]]:
            if PROPERTYTYPE(static_dynamic) == PROPERTYTYPE.STATIC:
                return [], None

            property_name = property_names[0]
            dates = self.grid_provider.dates_for_dynamic_property(
                property_name=property_name
            )
            dates = dates if dates else []
            current_date_options = current_date_options if current_date_options else []
            if set(dates) == {dateopt["value"] for dateopt in current_date_options}:
                return no_update, no_update
            normaldates = [datestr for datestr in dates if "_" not in datestr]
            diffdates = [datestr for datestr in dates if "_" in datestr]
            dates = normaldates + diffdates
            return (
                ([{"label": prop, "value": prop} for prop in dates]),
                [dates[0]] if dates else None,
            )



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class WellOverviewFilters(SettingsGroupABC):
    class Ids(StrEnum):
        SELECTED_WELLS = "selected-wells"

    def __init__(self, wells: List[str]) -> None:
        super().__init__("Filters")
        self._wells = wells

    def layout(self) -> List[Component]:
        return [
            wcc.SelectWithLabel(
                label="Well",
                size=min(10, len(self._wells)),
                id=self.register_component_unique_id(
                    WellOverviewFilters.Ids.SELECTED_WELLS
                ),
                options=[{"label": well, "value": well} for well in self._wells],
                value=self._wells,
                multi=True,
            )
        ]



import abc
import datetime
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Sequence

import pandas as pd


class Frequency(Enum):
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    QUARTERLY = "quarterly"
    YEARLY = "yearly"

    @classmethod
    def from_string_value(cls, value: str) -> Optional["Frequency"]:
        try:
            return cls(value)
        except ValueError:
            return None


@dataclass(frozen=True)
class VectorMetadata:
    unit: str
    is_total: bool
    is_rate: bool
    is_historical: bool
    keyword: str
    wgname: Optional[str]
    get_num: Optional[int]


# Class provides data for ensemble summary (timeseries)
class EnsembleSummaryProvider(abc.ABC):
    @abc.abstractmethod
    def vector_names(self) -> List[str]:
        """Returns list of all available vector names."""

    @abc.abstractmethod
    def vector_names_filtered_by_value(
        self,
        exclude_all_values_zero: bool = False,
        exclude_constant_values: bool = False,
    ) -> List[str]:
        """Returns list vector names with the option of excluding vectors that have only
        0-values and/or vectors where all the values are equal.
        """

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        """Returns list of all available realization numbers."""

    @abc.abstractmethod
    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        """Returns metadata for the specified vector. Returns None if no metadata
        exists or if any of the non-optional properties of `VectorMetadata` are missing.
        """

    @abc.abstractmethod
    def supports_resampling(self) -> bool:
        """Returns True if this provider supports resampling, otherwise False.
        A provider that doesn't support resampling will only accept None as value for
        the resampling_frequency parameter in `dates()` and `get_vectors_df()`.
        """

    @abc.abstractmethod
    def dates(
        self,
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> List[datetime.datetime]:
        """Returns the intersection of available dates.
        Note that when resampling_frequency is None, the pure intersection of the
        stored raw dates will be returned. Thus the returned list of dates will not include
        dates from long running realizations.
        For other resampling frequencies, the date range will be expanded to cover the entire
        time range of all the requested realizations before computing the resampled dates.
        """

    @abc.abstractmethod
    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        """Returns a Pandas DataFrame with data for the vectors specified in `vector_names.`

        For a provider that supports resampling, the `resampling_frequency` parameter
        controls the sampling frequency of the returned data. If `resampling_frequency` is
        None, the data will be returned with full/raw resolution.
        For a provider that does not support resampling, the `resampling_frequency` parameter
        must always be None, otherwise an exception will be raised.

        The returned DataFrame will always contain a 'DATE' and 'REAL' column in addition
        to columns for all the requested vectors.
        """

    @abc.abstractmethod
    def get_vectors_for_date_df(
        self,
        date: datetime.datetime,
        vector_names: Sequence[str],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        """Returns a Pandas DataFrame with data for the specified `date` and the vectors
        specified in `vector_names.`

        For a provider that supports resampling, all vectors will be resampled at the
        specified `date.` For providers that do not support resampling, an exact match on
        `date` will be required.

        The returned DataFrame will always contain a 'REAL' column in addition to
        columns for all the requested vectors.
        """



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
########################################

from typing import Callable, Optional, Tuple, Union

import numpy as np

# opm is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from opm.io.ecl import EclFile
except ImportError:
    pass

from ..eclipse_unit import ConvertUnits, CreateUnitConverter, EclUnitEnum, EclUnits
from .pvt_common import (
    EclPhaseIndex,
    EclPropertyTableRawData,
    FluidImplementation,
    InitFileDefinitions,
    PVDx,
    PVTx,
    PvxOBase,
    surface_mass_density,
)


class WetGas(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data for wet gas

    Attributes:
        interpolant: The interpolant object
    """

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        surface_mass_density_oil: float,
        surface_mass_density_gas: float,
        convert: Tuple[
            Callable[
                [
                    float,
                ],
                float,
            ],
            ConvertUnits,
        ],
    ) -> None:
        """Initializes a WetGas object.

        Creates an interpolant for wet gas for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data object
            surface_mass_density_oil: Surface mass density of oil
            surface_mass_density_gas: Surface mass density of gas
            convert: Tuple holding a callable and a ConvertUnits object for unit conversions

        """
        self.__interpolant = PVTx(index_table, raw, convert)
        self.__surface_mass_density_oil = surface_mass_density_oil
        self.__surface_mass_density_gas = surface_mass_density_gas

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Computes formation volume factor values
        for the given ratio and pressure values.

        Args:
            ratio: Ratio (key) values the volume factor values are requested for.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            Volume factor values for the given ratio and pressure values.

        """
        # Remember:
        # PKey   Inner   C0     C1         C2           C3
        # Pg     Rv      1/B    1/(B*mu)   d(1/B)/dRv   d(1/(B*mu))/dRv
        #        :       :      :          :            :
        return self.__interpolant.formation_volume_factor(pressure, ratio)

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given ratio and pressure values.

        Args:
            ratio: Ratio (key) values the viscosity values are requested for.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            Viscosity values for the given ratio and pressure values.

        """
        # Remember:
        # PKey   Inner   C0     C1         C2           C3
        # Pg     Rv      1/B    1/(B*mu)   d(1/B)/dRv   d(1/(B*mu))/dRv
        #        :       :      :          :            :
        return self.__interpolant.viscosity(pressure, ratio)

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the density values are requested for.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_g = (rho_g,sc + Rv * rho_o,sc) / B_g
        fvf_gas = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                (
                    self.__surface_mass_density_gas
                    + ratio_sample * self.__surface_mass_density_oil
                )
                / fvf_gas_sample
                for ratio_sample, fvf_gas_sample in zip(ratio, fvf_gas)
            ]
        )

    def get_keys(self) -> np.ndarray:
        """Returns all primary pressure values (Pg)"""
        return self.__interpolant.get_keys()

    def get_independents(self) -> np.ndarray:
        """Returns all gas ratio values (Rv)"""
        return self.__interpolant.get_independents()


class DryGas(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data for dry gas

    Attributes:
        interpolant: The interpolant object
    """

    def __init__(
        self,
        table_index: int,
        raw: EclPropertyTableRawData,
        surface_mass_density_gas: float,
        convert: ConvertUnits,
    ) -> None:
        """Initializes a DryGas object.

        Creates an interpolant for dry gas for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data object
            surface_mass_density_gas: Surface mass density of gas
            convert: ConvertUnits object for unit conversions

        """
        self.__interpolant = PVDx(table_index, raw, convert)
        self.__surface_mass_density_gas = surface_mass_density_gas

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Computes formation volume factor values
        for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            Volume factor values for the given pressure values.

        """
        return self.__interpolant.formation_volume_factor(pressure)

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            Viscosity values for the given pressure values.

        """
        return self.__interpolant.viscosity(pressure)

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_g = rho_g,sc / B_g
        fvf_gas = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                self.__surface_mass_density_gas / fvf_gas_sample
                for fvf_gas_sample in fvf_gas
            ]
        )

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is dry gas, there is no dependency on Rv.
        Hence, this method returns a list holding floats of value 0.0
        for each independent value.

        """
        return self.__interpolant.get_keys()

    def get_independents(self) -> np.ndarray:
        """Returns all independent pressure values (Pg)"""
        return self.__interpolant.get_independents()


class Gas(FluidImplementation):
    """Class for storing PVT tables for gas

    Holds a list of regions (one per PVT table).

    Attributes:
        surface_mass_densities: Surface mass densities
        keep_unit_system: True if the original unit system was kept
        original_unit_system: An ErtEclUnitEnum representing the original unit system
    """

    def __init__(
        self,
        raw: EclPropertyTableRawData,
        unit_system: int,
        surface_mass_densities: Tuple[np.ndarray, np.ndarray],
        keep_unit_system: bool = False,
    ) -> None:
        """Initializes a Gas object.

        Args:
            raw: Eclipse raw data object
            unit_system: The original unit system
            surface_mass_densities: Surface mass densities
            keep_unit_system:
                True if the original unit system shall be kept,
                False if units shall be converted to SI units.

        """
        super().__init__(keep_unit_system)
        self.surface_mass_densities = surface_mass_densities
        self.original_unit_system = EclUnitEnum(unit_system)
        self.create_pvt_interpolants(raw, unit_system)

    def formation_volume_factor_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the formation volume factor.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the formation volume factor.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return (
                rf"${{{unit_system.reservoir_volume().symbol}}}"
                rf"/{{{unit_system.surface_volume_gas().symbol}}}$"
            )
        return (
            f"{unit_system.reservoir_volume().symbol}"
            f"/{unit_system.surface_volume_gas().symbol}"
        )

    def ratio_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the phase ratio.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the phase ratio.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return (
                rf"${{{unit_system.surface_volume_liquid().symbol}}}/"
                rf"{{{unit_system.surface_volume_gas().symbol}}}$"
            )
        return (
            f"{unit_system.surface_volume_liquid().symbol}/"
            f"{unit_system.surface_volume_gas().symbol}"
        )

    def create_pvt_interpolants(
        self, raw: EclPropertyTableRawData, unit_system: int
    ) -> None:
        """Creates interpolants for PVT data based on the type of the gas (i.e. dry or wet gas).

        Args:
            raw: Eclipse raw data object
            unit_system: Number describing the unit system the values in the raw data are stored in

        Raises:
            A FluidImplementation.InvalidArgument exception
            in case of a size mismatch in the raw data.

        """
        if raw.num_primary == 0:
            raise super().InvalidArgument("Gas PVT table without primary lookup key")
        if raw.num_cols != 5:
            raise super().InvalidArgument("PVT table for gas must have five columns")
        if len(raw.primary_key) != (raw.num_primary * raw.num_tables):
            raise super().InvalidArgument(
                "Size mismatch in Pressure nodes of PVT table for gas"
            )
        if len(raw.data) != (
            raw.num_primary * raw.num_rows * raw.num_cols * raw.num_tables
        ):
            raise super().InvalidArgument(
                "Size mismatch in Condensed table data of PVT table for gas"
            )

        if raw.num_primary == 1:
            self.create_dry_gas(raw, unit_system)
        else:
            self.create_wet_gas(raw, unit_system)

    def dry_gas_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> ConvertUnits:
        """Creates a ConvertUnits object for unit conversions for dry gas.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            ConvertUnits object for unit conversions.

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvdx_unit_converter() or ConvertUnits(
            CreateUnitConverter.ToSI.pressure(unit_system),
            [
                CreateUnitConverter.ToSI.recip_fvf_gas(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_gas_visc(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_gas_deriv_press(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_gas_visc_deriv_press(unit_system),
            ],
        )

    def wet_gas_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> Tuple[Callable[[float,], float,], ConvertUnits]:
        """Creates a tuple consisting of a callable and a ConvertUnits object
        for unit conversions for wet gas.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            Tuple consisting of callable and ConvertUnits object

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvtx_unit_converter() or (
            CreateUnitConverter.ToSI.pressure(unit_system),
            ConvertUnits(
                CreateUnitConverter.ToSI.vaporised_oil_gas_ratio(unit_system),
                [
                    CreateUnitConverter.ToSI.recip_fvf_gas(unit_system),
                    CreateUnitConverter.ToSI.recip_fvf_gas_visc(unit_system),
                    CreateUnitConverter.ToSI.recip_fvf_gas_deriv_vap_oil(unit_system),
                    CreateUnitConverter.ToSI.recip_fvf_gas_visc_deriv_vap_oil(
                        unit_system
                    ),
                ],
            ),
        )

    def create_dry_gas(self, raw: EclPropertyTableRawData, unit_system: int) -> None:
        """Creates interpolants for dry gas from the given raw Eclipse data and uses
        a dry gas unit converter based on the given unit system.

        Args:
            raw: Eclipse raw data object
            unit_system: Integer representation of the unit system used in Eclipse data

        """
        cvrt = self.dry_gas_unit_converter(unit_system)

        self._regions = self.make_interpolants_from_raw_data(
            # Inner   C0     C1         C2           C3
            # Pg      1/B    1/(B*mu)   d(1/B)/dRv   d(1/(B*mu))/dRv
            #         :       :         :            :
            raw,
            lambda table_index, raw: DryGas(
                table_index, raw, self.surface_mass_densities[1][table_index], cvrt
            ),
        )

    def create_wet_gas(self, raw: EclPropertyTableRawData, unit_system: int) -> None:
        """Creates interpolants for wet gas from the given raw Eclipse data and uses
        a wet gas unit converter based on the given unit system.

        Args:
            raw: Eclipse raw data object
            unit_system: Integer representation of the unit system used in Eclipse data

        """
        cvrt = self.wet_gas_unit_converter(unit_system)

        self._regions = self.make_interpolants_from_raw_data(
            # PKey   Inner   C0     C1         C2           C3
            # Pg     Rv      1/B    1/(B*mu)   d(1/B)/dRv   d(1/(B*mu))/dRv
            #        :       :      :          :            :
            raw,
            lambda table_index, raw: WetGas(
                table_index,
                raw,
                self.surface_mass_densities[0][table_index],
                self.surface_mass_densities[1][table_index],
                cvrt,
            ),
        )

    def is_wet_gas(self) -> bool:
        """Checks if this fluid is wet gas.

        Returns:
            True if wet gas, else False

        """
        if len(self._regions) > 0:
            return isinstance(self._regions[0], WetGas)
        return False

    def is_dry_gas(self) -> bool:
        """Checks if this fluid is dry gas.

        Returns:
            True if dry gas, else False

        """
        if len(self._regions) > 0:
            return isinstance(self._regions[0], DryGas)
        return False

    @staticmethod
    def from_ecl_init_file(
        ecl_init_file: EclFile, keep_unit_system: bool = True
    ) -> Optional["Gas"]:
        """Reads the given Eclipse file and creates a Gas object from its data.

        Args:
            ecl_init_file: Eclipse INIT file
            keep_unit_system:
                Set to True if the unit system used in the Eclipse file
                shall be kept, False if SI shall be used.

        Returns:
            A Gas object or None if the data in the Eclipse file was invalid

        """
        intehead = ecl_init_file.__getitem__(InitFileDefinitions.INTEHEAD_KW)
        intehead_phase = intehead[InitFileDefinitions.INTEHEAD_PHASE_INDEX]

        if (intehead_phase & (1 << 2)) == 0:
            return None

        raw = EclPropertyTableRawData()

        tab_dims = ecl_init_file.__getitem__("TABDIMS")
        tab = ecl_init_file.__getitem__("TAB")

        num_rv = tab_dims[InitFileDefinitions.TABDIMS_NRPVTG_ITEM]
        num_pg = tab_dims[InitFileDefinitions.TABDIMS_NPPVTG_ITEM]

        raw.num_cols = 5
        raw.num_tables = tab_dims[InitFileDefinitions.TABDIMS_NTPVTG_ITEM]

        if raw.num_tables == 0:
            return None

        logihead = ecl_init_file.__getitem__(InitFileDefinitions.LOGIHEAD_KW)

        if logihead[InitFileDefinitions.LOGIHEAD_RV_INDEX]:
            raw.num_primary = num_pg
            raw.num_rows = num_rv

        else:
            raw.num_primary = 1
            raw.num_rows = num_pg

        num_tab_elements = raw.num_primary * raw.num_tables
        start = tab_dims[InitFileDefinitions.TABDIMS_JBPVTG_OFFSET_ITEM] - 1
        raw.primary_key = tab[start : start + num_tab_elements]

        num_tab_elements = (
            raw.num_primary * raw.num_rows * raw.num_cols * raw.num_tables
        )
        start = tab_dims[InitFileDefinitions.TABDIMS_IBPVTG_OFFSET_ITEM] - 1
        raw.data = tab[start : start + num_tab_elements]

        surface_mass_densities = (
            surface_mass_density(ecl_init_file, EclPhaseIndex.LIQUID, keep_unit_system),
            surface_mass_density(ecl_init_file, EclPhaseIndex.VAPOUR, keep_unit_system),
        )

        return Gas(
            raw,
            intehead[InitFileDefinitions.INTEHEAD_UNIT_INDEX],
            surface_mass_densities,
            keep_unit_system,
        )



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
#  Copyright (C) 2009-2012 SINTEF ICT, Applied Mathematics.
#
########################################

import re
from typing import Any, Dict, Union


class UnitBase:
    """This is a pseudo-base-class necessary due to class Prefix requiring
    a class type for operator type definitions.
    """

    def __init__(self, value: float, symbol: str) -> None:
        raise NotImplementedError("This pseudo-class must not be used directly.")

    @property
    def value(self) -> float:
        raise NotImplementedError("This pseudo-class must not be used directly.")

    @property
    def raw_symbol(self) -> str:
        raise NotImplementedError("This pseudo-class must not be used directly.")

    @property
    def symbol(self) -> str:
        raise NotImplementedError("This pseudo-class must not be used directly.")

    def __mul__(self, other: Any) -> Any:
        raise NotImplementedError("This pseudo-class must not be used directly.")

    def __truediv__(self, other: Any) -> Any:
        raise NotImplementedError("This pseudo-class must not be used directly.")


# pylint: disable=too-few-public-methods
class Prefix:
    """Namespace for unit prefixes.

    Attributes:
        micro
        milli
        centi
        deci
        kilo
        mega
        giga
    """

    class Base:
        """Holds prefix factor and symbol of a Prefix and allows multiplication with Units.

        Raises NotImplementedErrors for other operations.

        Attributes:
            factor: The prefix factor (e.g. for kilo: 1.0e3)
            symbol: The prefix symbol (e.g. for kilo: k)
        """

        def __init__(self, factor: float, symbol: str) -> None:
            """Initializes a Prefix with the given factor and symbol.

            Args:
                factor: The prefix factor (e.g. for kilo: 1.0e3)
                symbol: The prefix symbol (e.g. for kilo: k)

            """
            self.factor = factor
            self.symbol = symbol

        def __mul__(self, unit: UnitBase) -> UnitBase:  # type: ignore[no-untyped-def]
            """Applies this factor to the given unit and returns the result as a new unit.

            Does also create a new symbol.

            Args:
                unit: The unit this prefix shall be applied on.

            Returns:
                A new unit as a result of the application of this prefix.

            """
            if issubclass(unit.__class__, UnitBase):
                return unit.__class__(
                    self.factor * unit.value, f"{self.symbol}{unit.symbol}"
                )

            raise TypeError("Can only be multiplied with a Unit.")

        def __add__(self, other):  # type: ignore[no-untyped-def]
            raise NotImplementedError("Prefixes can only be multiplied with a Unit.")

        def __sub__(self, other):  # type: ignore[no-untyped-def]
            raise NotImplementedError("Prefixes can only be multiplied with a Unit.")

        def __truediv__(self, other):  # type: ignore[no-untyped-def]
            raise NotImplementedError("Prefixes can only be multiplied with a Unit.")

    micro = Base(1.0e-6, "\u00B5")
    milli = Base(1.0e-3, "m")
    centi = Base(1.0e-2, "c")
    deci = Base(1.0e-1, "d")
    kilo = Base(1.0e3, "k")
    mega = Base(1.0e6, "M")
    giga = Base(1.0e9, "G")


class Unit:
    """Namespace for units"""

    class Base(UnitBase):
        """Common class for all units.

        Used to have one object holding both value and symbol of the unit.
        Does also allow multiplication and division with other units and takes
        care of keeping its symbol tidy.
        """

        # pylint: disable=super-init-not-called
        def __init__(
            self,
            value: Union[float, UnitBase],  # type: ignore[name-defined]
            symbol: str,
        ) -> None:
            """Creates a new unit instance.

            Tidies the given symbol.

            Args:
                value: Either a value or another unit
                symbol: The symbol of the unit

            Returns:
                A new unit with the given value and a tidy symbol.

            """
            if isinstance(value, float):
                self.__value = value
            else:
                self.__value = value.value

            self.__symbol = symbol

            self.__tidy_symbol()

        @property
        def value(self) -> float:
            """
            Returns:
                The value of the unit.

            """
            return self.__value

        @property
        def raw_symbol(self) -> str:
            """
            Returns:
                The raw symbol of the unit.

            """
            return self.__symbol

        @property
        def symbol(self) -> str:
            """
            Returns:
                The formatted symbol of the unit.

            """
            return self.__formatted_symbol()

        def __formatted_symbol(self) -> str:
            powers = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
            symbol = self.__symbol.replace("^", "").translate(powers)
            symbol = re.sub(r"\(([^*+\-\/]+)\)", r"\1", symbol)
            return symbol

        def __tidy_symbol(self) -> None:
            # pylint: disable=too-many-branches
            # pylint: disable=too-many-statements
            def add_symbol(
                current_symbol: str,
                current_power: str,
                in_denominator: bool,
                numerator: Dict[str, int],
                denominator: Dict[str, int],
            ) -> None:
                if current_symbol != "":
                    if not in_denominator:
                        if current_symbol not in numerator:
                            numerator[current_symbol] = 0
                        numerator[current_symbol] += int(
                            current_power if current_power != "" else 1
                        )
                    else:
                        if current_symbol not in denominator:
                            denominator[current_symbol] = 0
                        denominator[current_symbol] += int(
                            current_power if current_power != "" else 1
                        )

            self.__symbol = "".join(self.__symbol.split())
            numerator: Dict[str, int] = {}
            denominator: Dict[str, int] = {}

            in_parantheses = False
            in_denominator = False
            power = False
            current_power = ""
            current_symbol = ""

            for i, char in enumerate(self.__symbol):
                if char in ["*", "/", "(", ")"]:
                    add_symbol(
                        current_symbol,
                        current_power,
                        in_denominator,
                        numerator,
                        denominator,
                    )

                if char in ["*", "/", "(", ")"]:
                    if char == "*":
                        if not in_parantheses:
                            in_denominator = False
                    elif char == "/":
                        in_denominator = True
                    elif char == "(":
                        in_parantheses = True
                    elif char == ")":
                        in_parantheses = False

                    current_symbol = ""
                    current_power = ""
                    power = False
                elif char == "^":
                    power = True
                else:
                    if power:
                        current_power += char
                    else:
                        current_symbol += char

                if i == len(self.__symbol) - 1:
                    add_symbol(
                        current_symbol,
                        current_power,
                        in_denominator,
                        numerator,
                        denominator,
                    )

            self.__symbol = ""

            for num, num_power in numerator.items():
                for denom, denom_power in denominator.items():
                    if num == denom:
                        denominator[num] = max(0, denom_power - num_power)
                        num_power = max(0, num_power - denom_power)
                        break

                if num_power > 0:
                    if self.__symbol == "":
                        self.__symbol += num
                    else:
                        self.__symbol += f"*{num}"
                    if num_power > 1:
                        self.__symbol += f"^{num_power}"

            if len(denominator) > 0:
                self.__symbol += "/("

                for denom, denom_power in denominator.items():
                    if denom_power > 0:
                        if self.__symbol[-1] == "(":
                            self.__symbol += denom
                        else:
                            self.__symbol += f"*{denom}"
                        if denom_power > 1:
                            self.__symbol += f"^{denom_power}"

                self.__symbol += ")"

        def __mul__(
            self,
            other: Union[UnitBase, Prefix.Base, float, int],
        ) -> UnitBase:
            """Multiplies this unit with the given unit/float/int and
            returns the result as a new unit.

            Does also create a new symbol.

            Args:
                unit: The unit/float/int to multiply with.

            Returns:
                A new unit as a result of the operation.

            """
            if issubclass(other.__class__, UnitBase):
                return self.__class__(
                    self.value * other.value, f"{self.raw_symbol}*{other.raw_symbol}"
                )
            if isinstance(other, float):
                return self.__class__(self.value * other, self.raw_symbol)
            if isinstance(other, int):
                return self.__class__(self.value * float(other), self.raw_symbol)

            raise TypeError(
                "You can only multiply this unit with another unit, a float or an integer."
            )

        def __truediv__(
            self,
            other: Union[UnitBase, Prefix.Base, float, int],
        ) -> UnitBase:
            """Divides this unit by the given unit/float/int and returns the result as a new unit.

            Does also create a new symbol.

            Args:
                unit: The unit/float/int to divide by.

            Returns:
                A new unit as a result of the operation.

            """
            if issubclass(other.__class__, UnitBase):
                return self.__class__(
                    self.value / other.value, f"{self.raw_symbol}/({other.raw_symbol})"
                )
            if isinstance(other, float):
                return self.__class__(self.value / other, self.raw_symbol)
            if isinstance(other, int):
                return self.__class__(self.value / float(other), self.raw_symbol)

            raise TypeError(
                "You can only divide this unit by another unit, a float or an integer."
            )

    # Common powers
    @staticmethod
    def square(unit: UnitBase) -> UnitBase:
        """Computes the square product of the given unit.

        Args:
            unit: The unit the square product shall be computed for.

        Returns:
            A new unit holding the square product of the given unit.

        """
        return unit * unit

    @staticmethod
    def cubic(unit: UnitBase) -> UnitBase:
        """Computes the cubic product of the given unit.

        Args:
            unit: The unit the cubic product shall be computed for.

        Returns:
            A new unit holding the cubic product of the given unit.

        """
        return unit * unit * unit

    #############################
    # Basic units and conversions
    #############################

    meter = Base(1.0, "m")

    inch = Base(Prefix.centi * meter * 2.54, "inch")

    feet = Base(inch * 12.0, "ft")

    # Time
    second = Base(1.0, "s")
    minute = Base(second * 60.0, "m")
    hour = Base(minute * 60.0, "h")
    day = Base(hour.value * 24.0, "d")
    # Reference: opm-common; might be changed later on to e,g, Gregorian 365.2524
    year = Base(day.value * 365.0, "y")

    # Volume
    gallon = Base(cubic.__func__(inch) * 231.0, "gal")  # type: ignore[attr-defined]
    stb = Base(gallon * 42.0, "stb")
    barrel = Base(gallon * 42.0, "bbl")
    liter = Base(cubic.__func__(Prefix.deci * meter), "l")  # type: ignore[attr-defined]

    # Mass
    kilogram = Base(1.0, "kg")
    gram = Base(kilogram * 1.0e-3, "g")
    pound = Base(kilogram * 0.45359237, "lb")

    # Energy
    joule = Base(1.0, "J")
    btu = Base(joule * 1054.3503, "BTU")

    # Standardised constant
    gravity = (meter * 9.80665) / square.__func__(second)  # type: ignore[attr-defined]

    ###############################
    # Derived units and conversions
    ###############################

    # Force
    newton = Base(kilogram * meter / square.__func__(second), "N")  # type: ignore[attr-defined]
    dyne = Base(newton * 1.0e-5, "dyn")
    lbf = Base(pound * gravity, "lbf")

    # Pressure
    pascal = Base(newton / square.__func__(meter), "Pa")  # type: ignore[attr-defined]
    bar = Base(pascal * 100000.0, "bar")  # pylint: disable=blacklisted-name
    atm = Base(pascal * 101325.0, "atm")
    psi = Base(lbf / square.__func__(inch), "psi")  # type: ignore[attr-defined]

    # Temperature
    deg_kelvin = Base(1.0, "K")

    deg_celsius = Base(1.0, "\u2103")
    deg_celsius_offset = 273.15

    deg_fahrenheit = Base(5.0 / 9.0, "\u2109")
    deg_fahrenheit_offset = 255.37

    # Viscosity
    pas = pascal * second
    poise = Base(Prefix.deci * pas, "P")

    # Permeability
    p_grad = atm / (Prefix.centi * meter)
    area = square.__func__(Prefix.centi * meter)  # type: ignore[attr-defined]
    flux = cubic.__func__(Prefix.centi * meter) / second  # type: ignore[attr-defined]
    velocity = flux / area
    visc = Base(Prefix.centi * poise, "cP")
    darcy = Base((velocity * visc) / p_grad, "D")

    class Convert:
        @staticmethod
        def from_(value: float, unit: float) -> float:
            return value * unit

        @staticmethod
        def to_(value: float, unit: float) -> float:
            return value / unit



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Sequence

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from ..._utils.perf_timer import PerfTimer
from ..ensemble_summary_provider._table_utils import (
    add_per_vector_min_max_to_table_schema_metadata,
    find_min_max_for_numeric_table_columns,
)
from ._field_metadata import create_column_metadata_from_field_meta
from .ensemble_table_provider import ColumnMetadata, EnsembleTableProvider

# Since PyArrow's actual compute functions are not seen by pylint
# pylint: disable=no-member


LOGGER = logging.getLogger(__name__)


def _sort_table_on_real(table: pa.Table) -> pa.Table:
    indices = pc.sort_indices(table, sort_keys=[("REAL", "ascending")])
    sorted_table = table.take(indices)
    return sorted_table


class EnsembleTableProviderImplArrow(EnsembleTableProvider):
    """This class implements a EnsembleTableProvider"""

    def __init__(self, arrow_file_name: Path) -> None:
        self._arrow_file_name = str(arrow_file_name)

        LOGGER.debug(f"init with arrow file: {self._arrow_file_name}")
        timer = PerfTimer()

        # We'll try and keep the file open for the life-span of the provider
        source = pa.memory_map(self._arrow_file_name, "r")
        et_open_ms = timer.lap_ms()

        self._cached_reader = pa.ipc.RecordBatchFileReader(source)
        et_create_reader_ms = timer.lap_ms()

        # Discover columns and realizations that are present in the arrow file
        column_names_on_file = self._cached_reader.schema.names
        self._column_names: List[str] = [
            colname
            for colname in column_names_on_file
            if colname not in ["REAL", "ENSEMBLE"]
        ]
        et_find_col_names_ms = timer.lap_ms()

        unique_realizations_on_file = (
            self._cached_reader.read_all().column("REAL").unique()
        )
        self._realizations: List[int] = unique_realizations_on_file.to_pylist()
        et_find_real_ms = timer.lap_ms()

        LOGGER.debug(
            f"init took: {timer.elapsed_s():.2f}s, "
            f"(open={et_open_ms}ms, create_reader={et_create_reader_ms}ms, "
            f"find_col_names={et_find_col_names_ms}ms, find_real={et_find_real_ms}ms), "
            f"#column_names={len(self._column_names)}, "
            f"#realization={len(self._realizations)}"
        )

    @staticmethod
    def write_backing_store_from_per_realization_tables(
        storage_dir: Path, storage_key: str, per_real_tables: Dict[int, pa.Table]
    ) -> None:
        # pylint: disable=too-many-locals
        @dataclass
        class Elapsed:
            concat_tables_s: float = -1
            build_add_real_col_s: float = -1
            sorting_s: float = -1
            find_and_store_min_max_s: float = -1
            write_s: float = -1

        elapsed = Elapsed()

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        LOGGER.debug(f"Writing backing store to arrow file: {arrow_file_name}")
        timer = PerfTimer()

        unique_column_names = set()
        for real_num, table in per_real_tables.items():
            unique_column_names.update(table.schema.names)

            if "REAL" in table.schema.names:
                raise ValueError(
                    f"Input tables should not have REAL column (real={real_num})"
                )

        LOGGER.debug(
            f"Concatenating {len(per_real_tables)} tables with "
            f"{len(unique_column_names)} unique column names"
        )

        full_table = pa.concat_tables(per_real_tables.values(), promote=True)
        elapsed.concat_tables_s = timer.lap_s()

        real_arr = np.empty(full_table.num_rows, np.int32)
        table_start_idx = 0
        for real_num, real_table in per_real_tables.items():
            real_arr[table_start_idx : table_start_idx + real_table.num_rows] = real_num
            table_start_idx += real_table.num_rows

        full_table = full_table.add_column(0, "REAL", pa.array(real_arr))
        elapsed.build_add_real_col_s = timer.lap_s()

        # Must sort table on real since interpolations work per realization
        # and we utilize slicing for speed
        full_table = _sort_table_on_real(full_table)
        elapsed.sorting_s = timer.lap_s()

        # Find per column min/max values and store them as metadata on table's schema
        per_vector_min_max = find_min_max_for_numeric_table_columns(full_table)
        full_table = add_per_vector_min_max_to_table_schema_metadata(
            full_table, per_vector_min_max
        )
        elapsed.find_and_store_min_max_s = timer.lap_s()

        # feather.write_feather(full_table, dest=arrow_file_name)
        with pa.OSFile(str(arrow_file_name), "wb") as sink:
            with pa.RecordBatchFileWriter(sink, full_table.schema) as writer:
                writer.write_table(full_table)
        elapsed.write_s = timer.lap_s()

        LOGGER.debug(
            f"Wrote backing store to arrow file in: {timer.elapsed_s():.2f}s ("
            f"concat_tables={elapsed.concat_tables_s:.2f}s, "
            f"build_add_real_col={elapsed.build_add_real_col_s:.2f}s, "
            f"sorting={elapsed.sorting_s:.2f}s, "
            f"find_and_store_min_max={elapsed.find_and_store_min_max_s:.2f}s, "
            f"write={elapsed.write_s:.2f}s)"
        )

    @staticmethod
    def write_backing_store_from_ensemble_dataframe(
        storage_dir: Path, storage_key: str, ensemble_df: pd.DataFrame
    ) -> None:

        table = pa.Table.from_pandas(ensemble_df, preserve_index=False)

        # The input DF may contain an ENSEMBLE column (which we'll drop before writing),
        # but it is probably an error if there is more than one unique value in it
        if "ENSEMBLE" in ensemble_df:
            if ensemble_df["ENSEMBLE"].nunique() > 1:
                raise KeyError("Input data contains more than one unique ensemble name")
            table = table.drop(["ENSEMBLE"])

        # Write to arrow format
        arrow_file_name: Path = storage_dir / (storage_key + ".arrow")
        with pa.OSFile(str(arrow_file_name), "wb") as sink:
            with pa.RecordBatchFileWriter(sink, table.schema) as writer:
                writer.write_table(table)

    @staticmethod
    def from_backing_store(
        storage_dir: Path, storage_key: str
    ) -> Optional["EnsembleTableProviderImplArrow"]:

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        if arrow_file_name.is_file():
            return EnsembleTableProviderImplArrow(arrow_file_name)

        return None

    def _get_or_read_schema(self) -> pa.Schema:
        if self._cached_reader:
            return self._cached_reader.schema

        source = pa.memory_map(self._arrow_file_name, "r")
        return pa.ipc.RecordBatchFileReader(source).schema

    def column_names(self) -> List[str]:
        return self._column_names

    def realizations(self) -> List[int]:
        return self._realizations

    def get_column_data(
        self, column_names: Sequence[str], realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:

        timer = PerfTimer()

        # For now guard against requesting the same column multiple times since that
        # will cause the conversion to pandas below to throw
        # This should probably raise an exception instead?
        if len(set(column_names)) != len(column_names):
            LOGGER.warning("The column_names argument contains duplicate names")
            column_names = list(dict.fromkeys(column_names))

        # We always want to include the the REAL column but watch out in case it is
        # already included in the column_names list
        columns_to_get = (
            ["REAL", *column_names] if "REAL" not in column_names else column_names
        )

        table = self._cached_reader.read_all().select(columns_to_get)
        et_read_ms = timer.lap_ms()

        if realizations:
            mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        df = table.to_pandas(ignore_metadata=True)
        et_to_pandas_ms = timer.lap_ms()

        LOGGER.debug(
            f"get_column_data() took: {timer.elapsed_ms()}ms "
            f"(read={et_read_ms}ms, filter={et_filter_ms}ms, to_pandas={et_to_pandas_ms}ms), "
            f"#cols={len(column_names)}, "
            f"#real={len(realizations) if realizations else 'all'}, "
            f"df.shape={df.shape}, file={Path(self._arrow_file_name).name}"
        )

        return df

    def column_metadata(self, column_name: str) -> Optional[ColumnMetadata]:
        schema = self._get_or_read_schema()
        field = schema.field(column_name)
        return create_column_metadata_from_field_meta(field)



import abc
from typing import List, Optional

import numpy as np
import xtgeo


class EnsembleGridProvider(abc.ABC):
    @abc.abstractmethod
    def provider_id(self) -> str:
        """Returns string ID of the provider."""

    @abc.abstractmethod
    def static_property_names(self) -> List[str]:
        """Returns list of all available static properties."""

    @abc.abstractmethod
    def dynamic_property_names(self) -> List[str]:
        """Returns list of all available dynamic properties."""

    @abc.abstractmethod
    def dates_for_dynamic_property(self, property_name: str) -> Optional[List[str]]:
        """Returns list of all available dates for a given dynamic property."""

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        """Returns list of all available realizations."""

    @abc.abstractmethod
    def get_3dgrid(
        self,
        realization: int,
    ) -> xtgeo.Grid:
        """Returns grid for specified realization"""

    @abc.abstractmethod
    def get_static_property_values(
        self, property_name: str, realization: int
    ) -> Optional[np.ndarray]:
        """Returns 1d cell values for a given static property"""

    @abc.abstractmethod
    def get_dynamic_property_values(
        self, property_name: str, property_date: str, realization: int
    ) -> Optional[np.ndarray]:
        """Returns 1d cell values for a given dynamic property"""



import datetime
from typing import Dict, List, Optional

from webviz_subsurface_components import ExpressionInfo

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._types import DeltaEnsemble
from .delta_ensemble_utils import (
    create_delta_ensemble_name_dict,
    create_delta_ensemble_provider_pair,
    is_delta_ensemble_providers_in_provider_set,
)
from .derived_vectors_accessor import (
    DerivedDeltaEnsembleVectorsAccessorImpl,
    DerivedEnsembleVectorsAccessorImpl,
    DerivedVectorsAccessor,
)


def create_derived_vectors_accessor_dict(
    ensembles: List[str],
    vectors: List[str],
    provider_set: EnsembleSummaryProviderSet,
    expressions: List[ExpressionInfo],
    delta_ensembles: List[DeltaEnsemble],
    resampling_frequency: Optional[Frequency],
    relative_date: Optional[datetime.datetime],
) -> Dict[str, DerivedVectorsAccessor]:
    """Create dictionary with ensemble name as key and derived vectors accessor
    as key.

    Obtain iterable object with ensemble name and corresponding vector data accessor.

    Creates derived vectors accessor based on ensemble type: Single ensemble or
    Delta ensemble.

    The derived vectors are based on listed vectors and created expressions.

    `Input:`
    * ensembles: List[str] - list of ensemble names
    * vectors List[str] - list of vectors to create accesses for
    * provider_set: EnsembleSummaryProviderSet - set of EnsembleSummaryProviders to
    obtain vector data
    * expressions: List[ExpressionInfo] - list of expressions for calculating vectors
    * delta_ensembles: List[DeltaEnsemble] - list of created delta ensembles
    * resampling_frequency: Optional[Frequency] - Resampling frequency setting for
    EnsembleSummaryProviders

    `Return:`
    * Dict[str, DerivedVectorsAccessor] - dictionary with ensemble name as key and
    DerivedVectorsAccessor implementations based on ensemble type - single ensemble
    or delta ensemble.

    TODO: Consider as a factory?
    """
    ensemble_data_accessor_dict: Dict[str, DerivedVectorsAccessor] = {}
    delta_ensemble_name_dict = create_delta_ensemble_name_dict(delta_ensembles)
    provider_names = provider_set.provider_names()
    for ensemble in ensembles:
        if ensemble in provider_names:
            ensemble_data_accessor_dict[ensemble] = DerivedEnsembleVectorsAccessorImpl(
                name=ensemble,
                provider=provider_set.provider(ensemble),
                vectors=vectors,
                expressions=expressions,
                resampling_frequency=resampling_frequency,
                relative_date=relative_date,
            )
        elif (
            ensemble in delta_ensemble_name_dict.keys()
            and is_delta_ensemble_providers_in_provider_set(
                delta_ensemble_name_dict[ensemble], provider_set
            )
        ):
            provider_pair = create_delta_ensemble_provider_pair(
                delta_ensemble_name_dict[ensemble], provider_set
            )
            ensemble_data_accessor_dict[
                ensemble
            ] = DerivedDeltaEnsembleVectorsAccessorImpl(
                name=ensemble,
                provider_pair=provider_pair,
                vectors=vectors,
                expressions=expressions,
                resampling_frequency=resampling_frequency,
                relative_date=relative_date,
            )

    return ensemble_data_accessor_dict



# pylint: disable=no-name-in-module
from webviz_config.plugins import BhpQc
from webviz_config.testing import WebvizComposite


def test_bhp_qc(_webviz_duo: WebvizComposite, shared_settings: dict) -> None:
    plugin = BhpQc(
        shared_settings["HM_SETTINGS"], ensembles=shared_settings["HM_ENSEMBLES"]
    )

    _webviz_duo.start_server(plugin)
    assert _webviz_duo.get_logs() == []



from typing import Callable, Dict, List, Union

import webviz_core_components as wcc
from dash import html


def plot_options_view(get_uuid: Callable, initial_layout: Dict) -> html.Div:
    return wcc.Selectors(
        label="Plot options",
        children=[
            wcc.Checklist(
                id=get_uuid("plotly_uirevision"),
                options=[{"label": "Keep plot range and zoom", "value": "keep"}],
            ),
            dropdown_for_plotly_layout(
                uuid=get_uuid("plotly_layout"),
                layout_attribute="xaxis_type",
                title="X-axis type",
                options=[
                    {"value": None, "label": "Automatic"},
                    {"value": "linear", "label": "Linear"},
                    {"value": "log", "label": "Log"},
                    {"value": "date", "label": "Date"},
                    {"value": "category", "label": "Category"},
                    {"value": "multicategory", "label": "Multicategory"},
                ],
                placeholder="automatic",
                value=initial_layout.get("xaxis", {}).get("type", None),
            ),
            dropdown_for_plotly_layout(
                uuid=get_uuid("plotly_layout"),
                layout_attribute="xaxis_autorange",
                title="X-axis direction",
                options=[
                    {"value": True, "label": "normal"},
                    {"value": "reversed", "label": "reversed"},
                ],
                value=initial_layout.get("xaxis", {}).get("autorange", True),
            ),
            dropdown_for_plotly_layout(
                uuid=get_uuid("plotly_layout"),
                layout_attribute="yaxis_type",
                title="Y-axis type",
                options=[
                    {"value": None, "label": "Automatic"},
                    {"value": "linear", "label": "Linear"},
                    {"value": "log", "label": "Log"},
                    {"value": "date", "label": "Date"},
                    {"value": "category", "label": "Category"},
                    {"value": "multicategory", "label": "Multicategory"},
                ],
                placeholder="automatic",
                value=initial_layout.get("yaxis", {}).get("type", None),
            ),
            dropdown_for_plotly_layout(
                uuid=get_uuid("plotly_layout"),
                layout_attribute="yaxis_autorange",
                title="Y-axis direction",
                options=[
                    {"value": True, "label": "normal"},
                    {"value": "reversed", "label": "reversed"},
                ],
                value=initial_layout.get("yaxis", {}).get("autorange", True),
            ),
        ],
    )


def dropdown_for_plotly_layout(
    uuid: str,
    layout_attribute: str,
    title: str,
    options: List[Dict],
    value: Union[List, str],
    placeholder: str = "Select...",
) -> html.Div:
    return wcc.Dropdown(
        label=title,
        id={
            "id": uuid,
            "layout_attribute": layout_attribute,
        },
        options=options,
        value=value,
        clearable=False,
        placeholder=placeholder,
    )



import json
import logging
from dataclasses import asdict, dataclass
from typing import Dict, Optional
from urllib.parse import quote

import flask
import geojson
import xtgeo
from dash import Dash

from .ensemble_fault_polygons_provider import (
    EnsembleFaultPolygonsProvider,
    FaultPolygonsAddress,
)

LOGGER = logging.getLogger(__name__)

_ROOT_URL_PATH = "/FaultPolygonsServer"

_FAULT_POLYGONS_SERVER_INSTANCE: Optional["FaultPolygonsServer"] = None


@dataclass(frozen=True)
class QualifiedAddress:
    provider_id: str
    address: FaultPolygonsAddress


class FaultPolygonsServer:
    def __init__(self, app: Dash) -> None:

        self._setup_url_rule(app)
        self._id_to_provider_dict: Dict[str, EnsembleFaultPolygonsProvider] = {}

    @staticmethod
    def instance(app: Dash) -> "FaultPolygonsServer":
        # pylint: disable=global-statement
        global _FAULT_POLYGONS_SERVER_INSTANCE
        if not _FAULT_POLYGONS_SERVER_INSTANCE:
            LOGGER.debug("Initializing FaultPolygonsServer instance")
            _FAULT_POLYGONS_SERVER_INSTANCE = FaultPolygonsServer(app)

        return _FAULT_POLYGONS_SERVER_INSTANCE

    def add_provider(self, provider: EnsembleFaultPolygonsProvider) -> None:

        provider_id = provider.provider_id()
        LOGGER.debug(f"Adding provider with id={provider_id}")

        existing_provider = self._id_to_provider_dict.get(provider_id)
        if existing_provider:
            # Issue a warning if there already is a provider registered with the same
            # id AND if the actual provider instance is different.
            # This should not be a problem, but will happen until the provider factory
            # gets caching.
            if existing_provider is not provider:
                LOGGER.warning(
                    f"Provider with id={provider_id} ignored, the id is already present"
                )
                return

        self._id_to_provider_dict[provider_id] = provider

    def encode_partial_url(
        self,
        provider_id: str,
        fault_polygons_address: FaultPolygonsAddress,
    ) -> str:
        if not provider_id in self._id_to_provider_dict:
            raise ValueError("Could not find provider")

        url_path: str = (
            f"{_ROOT_URL_PATH}/{quote(provider_id)}"
            f"/{quote(json.dumps(asdict(fault_polygons_address)))}"
        )

        return url_path

    def _setup_url_rule(self, app: Dash) -> None:
        @app.server.route(_ROOT_URL_PATH + "/<provider_id>/<fault_polygons_address>")
        def _handle_fault_polygons_request(
            provider_id: str,
            fault_polygons_address: str,
        ) -> flask.Response:
            LOGGER.debug(
                f"Handling fault_polygons_request: "
                f"full_fault_polygons_address={fault_polygons_address} "
            )

            fault_polygons_geojson = None
            # try:

            address = FaultPolygonsAddress(**json.loads(fault_polygons_address))
            provider = self._id_to_provider_dict[provider_id]
            fault_polygons = provider.get_fault_polygons(address)
            if fault_polygons is not None:
                fault_polygons_geojson = _create_fault_polygons_geojson(
                    polygons=fault_polygons
                )

            # except Exception as e:
            #     LOGGER.error("Error decoding fault polygons address")
            #     print(e)
            #     # flask.abort(404)
            featurecoll = (
                fault_polygons_geojson
                if fault_polygons_geojson is not None
                else {
                    "type": "FeatureCollection",
                    "features": [],
                }
            )

            return flask.Response(
                geojson.dumps(featurecoll), mimetype="application/geo+json"
            )


def _create_fault_polygons_geojson(polygons: xtgeo.Polygons) -> Dict:
    feature_arr = []
    for name, polygon in polygons.dataframe.groupby("POLY_ID"):
        coords = [list(zip(polygon.X_UTME, polygon.Y_UTMN))]
        feature = geojson.Feature(
            geometry=geojson.Polygon(coords),
            properties={"name": f"id:{name}", "color": [0, 0, 0, 255]},
        )
        feature_arr.append(feature)
    return geojson.FeatureCollection(features=feature_arr)



import json
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .._datainput.fmu_input import load_ensemble_set, load_parameters


class RunningTimeAnalysisFMU(WebvizPluginABC):
    """Can e.g. be used to investigate which jobs that are important for the running
time of realizations, and if specific parameter combinations increase running time or chance of
realization failure. Systematic realization failure could introduce bias to assisted history
matching.

Visualizations:
* Running time matrix, a heatmap of job running times relative to:
    * Same job in ensemble
    * Slowest job in ensemble
    * Slowest job in realization
* Parameter parallel coordinates plot:
    * Analyze running time and successful/failed run together with input parameters.

---

* **`ensembles`:** Which ensembles in `shared_settings` to include in check. Only required input.
* **`filter_shorter`:** Filters jobs with maximum run time in ensemble less than X seconds \
    (default: 10). Can be checked on/off interactively, this only sets the filtering value.
* **`status_file`:** Name of json file local per realization with job status \
    (default: `status.json`).
* **`visual_parameters`:** List of default visualized parameteres in parallel coordinates plot \
    (default: all parameters).

---

Parameters are picked up automatically from `parameters.txt` in individual realizations in
defined ensembles using `fmu-ensemble`.

The `status.json` file is the standard status file when running
[`ERT`](https://github.com/Equinor/ert) runs. If defining a different name, it still has to be
on the same format [(example file)](https://github.com/equinor/webviz-subsurface-testdata/\
blob/master/reek_history_match/realization-0/iter-0/status.json).
"""

    COLOR_MATRIX_BY_LABELS = [
        "Same job in ensemble",
        "Slowest job in realization",
        "Slowest job in ensemble",
    ]

    COLOR_PARCOORD_BY_LABELS = [
        "Successful/failed realization",
        "Running time of realization",
    ]

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        filter_shorter: Union[int, float] = 10,
        status_file: str = "status.json",
        visual_parameters: Optional[list] = None,
    ):
        super().__init__()
        self.filter_shorter = filter_shorter
        self.ens_paths = {
            ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
            for ens in ensembles
        }
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.ensembles = ensembles
        self.status_file = status_file
        self.parameter_df = load_parameters(
            ensemble_paths=self.ens_paths,
            ensemble_set_name="EnsembleSet",
            filter_file=None,
        )
        all_data_df = make_status_df(
            self.ens_paths, self.status_file
        )  # Has to be stored in one df due to webvizstore, see issue #206 in webviz-config
        self.job_status_df = all_data_df.loc["job"]
        self.real_status_df = all_data_df.loc["real"]
        self.visual_parameters = (
            visual_parameters if visual_parameters else self.parameters
        )
        self.set_callbacks(app)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("mode"),
                "content": (
                    "Switch between job running time matrix and parameter parallel coordinates."
                ),
            },
            {
                "id": self.uuid("ensemble"),
                "content": ("Display the realizations from the selected ensemble. "),
            },
            {
                "id": self.uuid("relative_runtime"),
                "content": ("Make the colorscale relative to the selected option."),
            },
        ]

    @property
    def parameters(self) -> List[str]:
        """Returns numerical input parameters"""
        return list(
            self.parameter_df.drop(["ENSEMBLE", "REAL"], axis=1)
            .apply(pd.to_numeric, errors="coerce")
            .dropna(how="all", axis="columns")
            .columns
        )

    @property
    def plot_fig(self) -> html.Div:
        return html.Div(
            style={
                "overflowX": "hidden",
                "width": "100%",
            },
            children=[
                html.Div(
                    id=self.uuid("plot_fig"),
                    style={"overflowX": "auto", "width": "100%"},
                    children=wcc.Graph(id=self.uuid("fig")),
                ),
                # Blank div: Makes sure that horizontal scrollbar is moved straight under the
                # figure instead of the figure div getting padded by whitespace down to height of
                # outer div.
                html.Div(style={"width": "100%"}),
            ],
        )

    @property
    def control_div(self) -> html.Div:
        return html.Div(
            children=[
                wcc.Selectors(
                    label="Mode",
                    children=[
                        wcc.RadioItems(
                            id=self.uuid("mode"),
                            options=[
                                {
                                    "label": "Running time matrix",
                                    "value": "running_time_matrix",
                                },
                                {
                                    "label": "Parameter parallel coordinates",
                                    "value": "parallel_coordinates",
                                },
                            ],
                            value="running_time_matrix",
                        ),
                    ],
                ),
                wcc.Selectors(
                    label="Ensemble",
                    children=[
                        wcc.Dropdown(
                            id=self.uuid("ensemble"),
                            options=[
                                {"label": ens, "value": ens} for ens in self.ensembles
                            ],
                            value=self.ensembles[0],
                            clearable=False,
                        ),
                    ],
                ),
                wcc.Selectors(
                    label="Coloring",
                    children=[
                        html.Div(
                            id=self.uuid("matrix_color"),
                            children=[
                                wcc.Dropdown(
                                    label="Color jobs relative to running time of:",
                                    id=self.uuid("relative_runtime"),
                                    options=[
                                        {"label": rel, "value": rel}
                                        for rel in RunningTimeAnalysisFMU.COLOR_MATRIX_BY_LABELS
                                    ],
                                    value=RunningTimeAnalysisFMU.COLOR_MATRIX_BY_LABELS[
                                        0
                                    ],
                                    clearable=False,
                                ),
                            ],
                        ),
                        html.Div(
                            id=self.uuid("parcoords_color"),
                            style={"display": "none"},
                            children=[
                                wcc.Dropdown(
                                    label="Color realizations relative to:",
                                    id=self.uuid("relative_real"),
                                    options=[
                                        {"label": rel, "value": rel}
                                        for rel in RunningTimeAnalysisFMU.COLOR_PARCOORD_BY_LABELS
                                    ],
                                    value=RunningTimeAnalysisFMU.COLOR_PARCOORD_BY_LABELS[
                                        0
                                    ],
                                    clearable=False,
                                ),
                            ],
                        ),
                    ],
                ),
                wcc.Selectors(
                    label="Filtering",
                    children=[
                        html.Div(
                            id=self.uuid("parameter_dropdown"),
                            style={"display": "none"},
                            children=[
                                wcc.SelectWithLabel(
                                    id=self.uuid("parameters"),
                                    style={"overflowX": "auto", "fontSize": "0.97rem"},
                                    options=[
                                        {"label": param, "value": param}
                                        for param in self.parameters
                                    ],
                                    multi=True,
                                    value=self.visual_parameters,
                                    size=min(50, len(self.visual_parameters)),
                                ),
                            ],
                        ),
                        html.Div(
                            id=self.uuid("filter_short_checkbox"),
                            children=[
                                wcc.Checklist(
                                    label="Filter jobs",
                                    id=self.uuid("filter_short"),
                                    options=[
                                        {
                                            "label": "Slowest in ensemble less than "
                                            f"{self.filter_shorter}s",
                                            "value": "filter_short",
                                        },
                                    ],
                                    value=["filter_short"],
                                ),
                            ],
                        ),
                    ],
                ),
            ]
        )

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            children=[
                wcc.Frame(
                    style={"flex": "1", "height": "90vh"}, children=self.control_div
                ),
                wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"flex": "6", "height": "90vh"},
                    children=self.plot_fig,
                ),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("fig"), "figure"),
            [
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("mode"), "value"),
                Input(self.uuid("relative_runtime"), "value"),
                Input(self.uuid("relative_real"), "value"),
                Input(self.uuid("parameters"), "value"),
                Input(self.uuid("filter_short"), "value"),
            ],
        )
        def _update_fig(
            ens: str,
            mode: str,
            rel_runtime: str,
            rel_real: str,
            params: Union[str, List[str]],
            filter_short: List[str],
        ) -> dict:
            """Update main figure
            Dependent on `mode` it will call rendering of the chosen form of visualization
            """
            if mode == "running_time_matrix" and "filter_short" in filter_short:
                return render_matrix(
                    self.job_status_df[
                        (self.job_status_df["ENSEMBLE"] == ens)
                        & (self.job_status_df["JOB_MAX_RUNTIME"] >= self.filter_shorter)
                    ],
                    rel_runtime,
                    self.plotly_theme,
                )
            if mode == "running_time_matrix":
                return render_matrix(
                    self.job_status_df[(self.job_status_df["ENSEMBLE"] == ens)],
                    rel_runtime,
                    self.plotly_theme,
                )

            # Otherwise: parallel coordinates
            # Ensure selected parameters is a list
            params = params if isinstance(params, list) else [params]
            # Color by success or runtime, for runtime drop unsuccesful
            colormap_labels: Union[List[str], None]
            if rel_real == "Successful/failed realization":
                plot_df = self.real_status_df[self.real_status_df["ENSEMBLE"] == ens]
                colormap = make_colormap(
                    self.plotly_theme["layout"]["colorway"], discrete=2
                )
                color_by_col = "STATUS_BOOL"
                colormap_labels = ["Failed", "Success"]
            else:
                plot_df = self.real_status_df[
                    (self.real_status_df["ENSEMBLE"] == ens)
                    & (self.real_status_df["STATUS_BOOL"] == 1)
                ]
                colormap = self.plotly_theme["layout"]["colorscale"]["sequential"]
                color_by_col = "RUNTIME"
                colormap_labels = None

            # Call rendering of parallel coordinate plot
            return render_parcoord(
                plot_df,
                params,
                self.plotly_theme,
                colormap,
                color_by_col,
                colormap_labels,
            )

        @app.callback(
            [
                Output(self.uuid("matrix_color"), "style"),
                Output(self.uuid("parcoords_color"), "style"),
                Output(self.uuid("parameter_dropdown"), "style"),
                Output(self.uuid("filter_short_checkbox"), "style"),
            ],
            [Input(self.uuid("mode"), "value")],
        )
        def _update_mode(mode: str) -> Tuple[dict, dict, dict, dict]:
            """Switch displayed mode between running time matrix and parallel coordinates"""
            if mode == "running_time_matrix":
                style = (
                    {"display": "block"},
                    {"display": "none"},
                    {"display": "none"},
                    {"display": "block"},
                )
            else:
                style = (
                    {"display": "none"},
                    {"display": "block"},
                    {"display": "block"},
                    {"display": "none"},
                )
            return style

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (
                make_status_df,
                [
                    {
                        "ens_paths": self.ens_paths,
                        "status_file": self.status_file,
                    }
                ],
            ),
            (
                load_parameters,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "filter_file": None,
                    },
                ],
            ),
        ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def render_matrix(status_df: pd.DataFrame, rel: str, theme: dict) -> dict:
    """Render matrix
    Returns figure object as heatmap for the chosen ensemble and scaling method.
    """
    if rel == "Same job in ensemble":
        z = list(status_df["JOB_SCALED_RUNTIME"])
    elif rel == "Slowest job in realization":
        z = list(status_df["REAL_SCALED_RUNTIME"])
    else:
        z = list(status_df["ENS_SCALED_RUNTIME"])
    data = {
        "type": "heatmap",
        "x": list(status_df["REAL"]),
        "y": list(status_df["JOB_ID"]),
        "z": z,
        "zmin": 0,
        "zmax": 1,
        "text": list(status_df["HOVERINFO"]),
        "hoverinfo": "text",
        "colorscale": theme["layout"]["colorscale"]["sequential"],
        "colorbar": {
            "tickvals": [
                0,
                0.5,
                1,
            ],
            "ticktext": [
                "0 %",
                "50 %",
                "100 %",
            ],
            "xanchor": "left",
        },
    }
    layout = {}
    layout.update(theme["layout"])
    layout.update(
        {
            "paper_bgcolor": "rgba(0,0,0,0)",
            "plot_bgcolor": "rgba(0,0,0,0)",
            "margin": {
                "t": 50,
                "b": 50,
                "l": 50,
            },
            "xaxis": {
                "ticks": "",
                "title": "Realizations",
                "showgrid": False,
                "side": "top",
            },
            "yaxis": {
                "ticks": "",
                "showticklabels": True,
                "tickmode": "array",
                "tickvals": list(status_df["JOB_ID"]),
                "ticktext": list(status_df["JOB"]),
                "showgrid": False,
                "automargin": True,
                "autorange": "reversed",
                "type": "category",
            },
            "height": max(350, len(status_df["JOB_ID"].unique()) * 15),
            "width": max(400, len(status_df["REAL"].unique()) * 12 + 250),
        }
    )

    return {"data": [data], "layout": layout}


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def render_parcoord(
    plot_df: pd.DataFrame,
    params: List[str],
    theme: dict,
    colormap: Union[List[str], List[list]],
    color_col: str,
    colormap_labels: Union[List[str], None] = None,
) -> dict:
    """Renders parallel coordinates plot"""
    # Create parcoords dimensions (one per parameter)
    dimensions = [
        {"label": param, "values": plot_df[param].values.tolist()} for param in params
    ]

    # Parcoords data dict
    data: dict = {
        "line": {
            "color": plot_df[color_col].values.tolist(),
            "colorscale": colormap,
            "showscale": True,
        },
        "dimensions": dimensions,
        "labelangle": -90,
        "labelside": "bottom",
        "type": "parcoords",
    }
    if color_col == "STATUS_BOOL":
        data["line"].update(
            {
                "cmin": -0.5,
                "cmax": 1.5,
                "colorbar": {
                    "tickvals": [0, 1],
                    "ticktext": colormap_labels,
                    "title": "Status",
                    "xanchor": "right",
                    "x": -0.02,
                    "len": 0.3,
                },
            },
        )
    else:
        data["line"].update(
            {
                "colorbar": {
                    "title": "Running time",
                    "xanchor": "right",
                    "x": -0.02,
                },
            },
        )

    layout = {}
    layout.update(theme["layout"])
    # Ensure sufficient spacing between each dimension and margin for labels
    width = len(dimensions) * 100 + 250
    margin_b = max([len(param) for param in params]) * 8
    layout.update({"width": width, "height": 800, "margin": {"b": margin_b, "t": 30}})
    return {"data": [data], "layout": layout}


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def make_status_df(
    ens_paths: dict,
    status_file: str,
) -> pd.DataFrame:
    """Return DataFrame of information from status.json files.
    *Finds status.json filepaths.
    For jobs:
    *Loads data into pandas DataFrames.
    *Calculates runtimes and normalized runtimes.
    *Creates hoverinfo column to be used in visualization.
    For realizations:
    *Creates DataFrame of success/failure and total running time.
    """

    parameter_df = load_parameters(
        ensemble_paths=ens_paths,
        ensemble_set_name="EnsembleSet",
        filter_file=None,
    )
    # sub-method to process ensemble data when all realizations in ensemble have been processed
    def ensemble_post_processing() -> list:
        # add missing realizations to get whitespace in heatmap matrix
        if len(set(range(min(reals), max(reals) + 1))) > len(set(reals)):
            missing_df = ens_dfs[0].copy()
            missing_df["STATUS"] = "Realization not started"
            missing_df["RUNTIME"] = np.NaN
            missing_df["JOB_SCALED_RUNTIME"] = np.NaN
            missing_df["ENS_SCALED_RUNTIME"] = np.NaN
            for missing_real in set(range(min(reals), max(reals) + 1)).difference(
                set(reals)
            ):
                ens_dfs.append(missing_df.copy())
                ens_dfs[-1]["REAL"] = missing_real
                ens_dfs[-1]["ENSEMBLE"] = ens
        # Concatenate realization DataFrames to an Ensemble DataFrame and store in list
        job_status_dfs.append(pd.concat(ens_dfs))
        # Find max running time of job in ensemble and create scaled columns
        job_status_dfs[-1]["JOB_MAX_RUNTIME"] = pd.concat(
            [ens_max_job_runtime] * (len(ens_dfs))
        )
        job_status_dfs[-1]["JOB_SCALED_RUNTIME"] = (
            job_status_dfs[-1]["RUNTIME"] / job_status_dfs[-1]["JOB_MAX_RUNTIME"]
        )
        job_status_dfs[-1]["ENS_SCALED_RUNTIME"] = job_status_dfs[-1][
            "RUNTIME"
        ] / np.amax(ens_max_job_runtime)
        # Return ensemble DataFrame list updated with the latest ensemble
        return job_status_dfs

    # find status filepaths
    ens_set = load_ensemble_set(ens_paths, filter_file=None)
    df = pd.concat(
        [
            ens_set[ens].find_files(status_file).assign(ENSEMBLE=ens)
            for ens in ens_set.ensemblenames
        ]
    )
    # Initial values for local variables
    job_status_dfs: list = []
    ens_dfs: list = []
    real_status: list = []
    ens_max_job_runtime = 1
    ens = ""
    reals: list = []

    # Loop through identified filepaths and get realization data
    for row in df.itertuples(index=False):
        # Load each json-file to a DataFrame for the realization
        with open(row.FULLPATH) as fjson:
            status_dict = json.load(fjson)
        real_df = pd.DataFrame(status_dict["jobs"])

        # If new ensemble, calculate ensemble scaled runtimes
        # for previous ensemble and reset temporary ensemble data
        if ens != row.ENSEMBLE:
            if ens == "":  # First ensemble
                ens = row.ENSEMBLE
            else:  # Store last ensemble and reset temporary ensemble data
                job_status_dfs = ensemble_post_processing()
                ens_max_job_runtime = 1
                ens_dfs = []
                ens = row.ENSEMBLE
                reals = []

        # Additional realization data into realization DataFrame
        real_df["RUNTIME"] = real_df["end_time"] - real_df["start_time"]
        real_df["REAL"] = row.REAL
        real_df["ENSEMBLE"] = row.ENSEMBLE
        real_df["REAL_SCALED_RUNTIME"] = real_df["RUNTIME"] / max(
            real_df["RUNTIME"].dropna()
        )
        real_df = real_df[
            ["ENSEMBLE", "REAL", "RUNTIME", "REAL_SCALED_RUNTIME", "name", "status"]
        ].rename(columns={"name": "JOB", "status": "STATUS"})
        # Status DataFrame to be used with parallel coordinates
        if all(real_df["STATUS"] == "Success"):
            real_status.append(
                {
                    "ENSEMBLE": row.ENSEMBLE,
                    "REAL": row.REAL,
                    "STATUS": "Success",
                    "STATUS_BOOL": 1,
                    "RUNTIME": status_dict["end_time"] - status_dict["start_time"],
                }
            )
        else:
            real_status.append(
                {
                    "ENSEMBLE": row.ENSEMBLE,
                    "REAL": row.REAL,
                    "STATUS": "Failure",
                    "STATUS_BOOL": 0,
                    "RUNTIME": None,
                }
            )

        # Need unique job ids names to separate jobs in same realization with same name in json file
        real_df["JOB_ID"] = range(0, len(real_df["JOB"]))

        # Update max runtime for jobs in ensemble
        ens_max_job_runtime = np.fmax(real_df["RUNTIME"], ens_max_job_runtime)

        # Append realization to ensemble data
        reals.append(row.REAL)
        ens_dfs.append(real_df)

    # Add last ensemble
    job_status_dfs = ensemble_post_processing()
    job_status_df = pd.concat(job_status_dfs, sort=False)

    # Create hoverinfo
    job_status_df["HOVERINFO"] = (
        "Real: "
        + job_status_df["REAL"].astype(str)
        + "<br>"
        + "Job: #"
        + job_status_df["JOB_ID"].astype(str)
        + "<br>"
        + job_status_df["JOB"].astype(str)
        + "<br>"
        + "Running time: "
        + job_status_df["RUNTIME"].astype(str)
        + " s"
        + "<br>"
        + "Status: "
        + job_status_df["STATUS"]
    )
    # Create dataframe of realization status and merge with realization parameters for parameter
    # parallel coordinates
    real_status_df = pd.DataFrame(real_status).merge(
        parameter_df, on=["ENSEMBLE", "REAL"]
    )
    # Has to be stored in one df due to webvizstore, see issue #206 in webviz-config
    return pd.concat([job_status_df, real_status_df], keys=["job", "real"], sort=False)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def make_colormap(color_array: list, discrete: int = None) -> list:
    """
    Returns a colormap:
    * If the `discrete` variable is set to an integer x, the colormap will be a discrete map of
    size x evenly sampled from the given color_array.
    * If discrete not defined or `None`: assumes continuous colormap and returns the given
    color_array.
    """
    if discrete is None:
        colormap = color_array
    else:
        colormap = []
        for i in range(0, discrete):
            colormap.append([i / discrete, color_array[i]])
            colormap.append([(i + 1) / discrete, color_array[i]])
    return colormap



# pylint: disable=no-name-in-module
from webviz_config.plugins import LinePlotterFMU


def test_line_plotter_fmu(dash_duo, app, testdata_folder, shared_settings) -> None:

    plugin = LinePlotterFMU(
        app,
        shared_settings["HM_SETTINGS"],
        aggregated_csvfile=testdata_folder
        / "reek_test_data"
        / "aggregated_data"
        / "smry_hm.csv",
        aggregated_parameterfile=testdata_folder
        / "reek_test_data"
        / "aggregated_data"
        / "parameters_hm.csv",
        observation_file=testdata_folder / "reek_test_data" / "observations.yml",
        observation_group="smry",
        remap_observation_values={"DATE": "date"},
        initial_data={
            "x": "DATE",
            "y": "FOPR",
            "ensembles": ["iter-0", "iter-3"],
            "colors": {"iter-0": "red", "iter-3": "blue"},
        },
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



import json
from pathlib import Path
from typing import Callable, Dict, List, Tuple
from uuid import uuid4

import numpy as np
import pandas as pd
import webviz_subsurface_components as wsc
from dash import html
from scipy.stats import chi2
from webviz_config import WebvizPluginABC, WebvizSettings

from .._datainput.history_match import extract_mismatch


class HistoryMatch(WebvizPluginABC):
    """Visualizes the quality of the history match.

---

* **`ensembles`:** List of the ensembles in `shared_settings` to visualize.
* **`observation_file`:** Path to the observation `.yaml` file \
(absolute or relative to config file).

---
Parameter values are extracted automatically from the `parameters.txt` files
of the individual realizations of your given `ensembles`, using the `fmu-ensemble` library.

?> The `observation_file` is a common (optional) file for all ensembles, which can be \
converted from e.g. ERT and ResInsight formats using the [fmuobs]\
(https://equinor.github.io/subscript/scripts/fmuobs.html) script. \
[An example of the format can be found here]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/share/\
observations/observations.yml).
"""

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: List[str],
        observation_file: Path,
    ):

        super().__init__()

        self.observation_file = observation_file

        self.ensembles = ensembles
        self.ens_paths = {
            ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
            for ens in ensembles
        }

        data = extract_mismatch(self.ens_paths, self.observation_file)
        self.hm_data = json.dumps(self._prepare_data(data))

        self.hm_id = f"hm-id-{uuid4()}"

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (
                extract_mismatch,
                [
                    {
                        "ens_paths": self.ens_paths,
                        "observation_file": self.observation_file,
                    }
                ],
            )
        ]

    def _prepare_data(self, data: pd.DataFrame) -> dict:
        data = data.copy().reset_index()

        num_obs_groups = len(data.obs_group_name.unique())

        data["avg_pos"] = data["total_pos"] / data["number_data_points"]
        data["avg_neg"] = data["total_neg"] / data["number_data_points"]

        iterations = []
        for ensemble in self.ensembles:
            df = data[data.ensemble_name == ensemble]
            iterations.append(df.groupby("obs_group_name").mean())

        sorted_iterations = HistoryMatch._sort_iterations(iterations)

        iterations_dict = HistoryMatch._iterations_to_dict(
            sorted_iterations, self.ensembles
        )

        confidence_sorted = _get_sorted_edges(num_obs_groups)
        confidence_unsorted = _get_unsorted_edges()

        data = {}
        data["iterations"] = iterations_dict
        data["confidence_interval_sorted"] = confidence_sorted
        data["confidence_interval_unsorted"] = confidence_unsorted

        return data

    @staticmethod
    def _sort_iterations(iterations: List[pd.DataFrame]) -> List[pd.DataFrame]:
        sorted_data = []

        for df in iterations:
            sorted_df = df.copy()

            sorted_data.append(
                sorted_df.assign(f=sorted_df["avg_pos"] + sorted_df["avg_neg"])
                .sort_values("f", ascending=False)
                .drop("f", axis=1)
            )

        return sorted_data

    @staticmethod
    def _iterations_to_dict(
        iterations: List[pd.DataFrame], labels: List[str]
    ) -> List[dict]:
        retval = []

        for iteration, label in zip(iterations, labels):
            retval.append(
                {
                    "name": label,
                    "positive": iteration["avg_pos"].tolist(),
                    "negative": iteration["avg_neg"].tolist(),
                    "labels": iteration.index.tolist(),
                }
            )

        return retval

    @property
    def layout(self) -> html.Div:
        return html.Div([wsc.HistoryMatch(id=self.hm_id, data=self.hm_data)])


def _get_unsorted_edges() -> dict:
    """P10 - P90 unsorted edge coordinates"""

    retval = {"low": chi2.ppf(0.1, 1), "high": chi2.ppf(0.9, 1)}

    return retval


def _get_sorted_edges(number_observation_groups: int) -> Dict[str, list]:
    """P10 - P90 sorted edge coordinates"""

    monte_carlo_iterations = 100000

    sorted_values = np.empty((number_observation_groups, monte_carlo_iterations))

    for i in range(monte_carlo_iterations):
        sorted_values[:, i] = np.sort(
            np.random.chisquare(df=1, size=number_observation_groups)
        )

    sorted_values = np.flip(sorted_values, 0)

    p10 = np.percentile(sorted_values, 90, axis=1)
    p90 = np.percentile(sorted_values, 10, axis=1)

    # Dictionary with two arrays (P10, P90). Each array of length equal
    # to number of observation groups i.e. number of items along y axis.
    # These values are to be used for drawing the stair stepped
    # sorted P10-P90 area:

    coordinates = {"low": list(p10), "high": list(p90)}

    return coordinates



import logging
import time

import numpy as np
import pyarrow as pa

from webviz_subsurface._providers.ensemble_summary_provider._resampling import (
    sample_segmented_multi_real_table_at_date,
)


def _create_table(
    num_reals: int, start_date: np.datetime64, end_date: np.datetime64, num_columns: int
) -> pa.Table:

    date_arr_np = np.empty(0, np.datetime64)
    real_arr_np = np.empty(0, np.int32)

    for real in range(0, num_reals):
        dates_for_this_real = np.arange(start_date, end_date + 1)
        dates_for_this_real = dates_for_this_real.astype("datetime64[ms]")
        real_arr_np = np.concatenate(
            (real_arr_np, np.full(len(dates_for_this_real), real))
        )
        date_arr_np = np.concatenate((date_arr_np, dates_for_this_real))

    print(
        f"real_arr_np (num unique={len(np.unique(real_arr_np))}  len={len(real_arr_np)}):"
    )
    print(real_arr_np)
    print(
        f"date_arr_np (num unique={len(np.unique(date_arr_np))}  len={len(date_arr_np)}):"
    )
    print(date_arr_np)

    field_list = []
    columndata_list = []
    field_list.append(pa.field("DATE", pa.timestamp("ms")))
    field_list.append(pa.field("REAL", pa.int64()))
    columndata_list.append(pa.array(date_arr_np))
    columndata_list.append(pa.array(real_arr_np))

    num_rows = len(real_arr_np)

    for colnum in range(0, num_columns):
        if (colnum % 2) == 0:
            metadata = {b"is_rate": b'{"is_rate": False}'}
        else:
            metadata = {b"is_rate": b'{"is_rate": True}'}

        field_list.append(pa.field(f"c_{colnum}", pa.float32(), metadata=metadata))

        valarr = np.linspace(colnum, colnum + num_rows, num_rows)
        columndata_list.append(pa.array(valarr))

    schema = pa.schema(field_list)
    return pa.table(columndata_list, schema=schema)


def main() -> None:
    print()
    print("## Running resampling performance tests")
    print("## =================================================")

    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(levelname)-3s [%(name)s]: %(message)s",
    )
    logging.getLogger("webviz_subsurface").setLevel(level=logging.INFO)
    logging.getLogger("webviz_subsurface").setLevel(level=logging.DEBUG)

    # table = _create_table(
    #     num_reals=3,
    #     start_date=np.datetime64("2020-12-30"),
    #     end_date=np.datetime64("2021-01-05"),
    #     num_columns=4,
    # )

    table = _create_table(
        num_reals=100,
        start_date=np.datetime64("2000-01-01", "M"),
        end_date=np.datetime64("2099-12-31", "M"),
        num_columns=10000,
    )

    print("## table shape (rows,columns):", table.shape)
    # print(table.to_pandas())

    start_tim = time.perf_counter()

    res = sample_segmented_multi_real_table_at_date(
        table, np.datetime64("2098-01-03", "ms")
    )

    # res = sample_segmented_multi_real_table_at_date(
    #     table, np.datetime64("2098-01-01", "ms")
    # )

    elapsed_time_ms = int(1000 * (time.perf_counter() - start_tim))

    # print(res)
    # print(res.to_pandas())

    print("## res shape:", res.shape)

    print(f"## sample at date took: {elapsed_time_ms}ms")


# Running:
#   python -m webviz_subsurface._providers.ensemble_summary_provider.dev_resampling_perf_testing
# -------------------------------------------------------------------------
if __name__ == "__main__":
    main()



# pylint: disable=no-name-in-module
from webviz_config.plugins import ReservoirSimulationTimeSeriesRegional


def test_reservoir_simulation_timeseries_regional(
    dash_duo, app, shared_settings, testdata_folder
) -> None:
    plugin = ReservoirSimulationTimeSeriesRegional(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        fipfile=testdata_folder
        / "01_drogon_ahm"
        / "realization-0"
        / "iter-0"
        / "share"
        / "results"
        / "tables"
        / "fip.yml",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from .ensemble_fault_polygons_provider import (
    EnsembleFaultPolygonsProvider,
    SimulatedFaultPolygonsAddress,
)
from .ensemble_fault_polygons_provider_factory import (
    EnsembleFaultPolygonsProviderFactory,
)
from .fault_polygons_server import FaultPolygonsAddress, FaultPolygonsServer



from typing import List

import webviz_core_components as wcc
from dash import html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import ChartType


class WellOverviewLayoutOptions(SettingsGroupABC):
    class Ids(StrEnum):
        CHARTTYPE_SETTINGS = "charttype-settings"
        CHARTTYPE_CHECKLIST = "charttype-checklist"

    def __init__(self) -> None:
        super().__init__("Layout Options")

    def layout(self) -> List[Component]:
        settings_id = self.register_component_unique_id(self.Ids.CHARTTYPE_SETTINGS)
        checklist_id = self.register_component_unique_id(self.Ids.CHARTTYPE_CHECKLIST)
        return [
            html.Div(
                children=[
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.BAR},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.BAR},
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {"label": "Overlay bars", "value": "overlay_bars"},
                                {
                                    "label": "Show prod as text",
                                    "value": "show_prod_text",
                                },
                                {
                                    "label": "White background",
                                    "value": "white_background",
                                },
                            ],
                            value=["legend"],
                        ),
                    ),
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.PIE},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.PIE},
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {
                                    "label": "Show prod as text",
                                    "value": "show_prod_text",
                                },
                            ],
                            value=[],
                        ),
                    ),
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.AREA},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.AREA},
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {
                                    "label": "White background",
                                    "value": "white_background",
                                },
                            ],
                            value=["legend"],
                        ),
                    ),
                ],
            ),
        ]



from .ensemble_grid_provider import EnsembleGridProvider
from .ensemble_grid_provider_factory import EnsembleGridProviderFactory
from .grid_viz_service import CellFilter, GridVizService, PickResult, PropertySpec, Ray



from typing import Any, Dict, List, Optional, Tuple

import dash
import webviz_core_components as wcc
from dash import Input, Output, State, callback, dcc, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._providers.ensemble_surface_provider.ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    SurfaceStatistic,
)
from webviz_subsurface.plugins._co2_leakage._utilities.callbacks import property_origin
from webviz_subsurface.plugins._co2_leakage._utilities.generic import MapAttribute


class ViewSettings(SettingsGroupABC):
    class Ids(StrEnum):
        FORMATION = "formation"
        ENSEMBLE = "ensemble"
        REALIZATION = "realization"

        PROPERTY = "property"
        STATISTIC = "statistic"
        COLOR_SCALE = "color-scale"
        CM_MIN = "cm-min"
        CM_MAX = "cm-max"
        CM_MIN_AUTO = "cm-min-auto"
        CM_MAX_AUTO = "cm-max-auto"

        PLUME_THRESHOLD = "plume-threshold"
        PLUME_SMOOTHING = "plume-smoothing"

    def __init__(
        self,
        ensemble_paths: Dict[str, str],
        ensemble_surface_providers: Dict[str, EnsembleSurfaceProvider],
        initial_surface: Optional[str],
        map_attribute_names: Dict[MapAttribute, str],
        color_scale_names: List[str],
    ):
        super().__init__("Settings")
        self._ensemble_paths = ensemble_paths
        self._ensemble_surface_providers = ensemble_surface_providers
        self._map_attribute_names = map_attribute_names
        self._color_scale_names = color_scale_names
        self._initial_surface = initial_surface

    def layout(self) -> List[Component]:
        return [
            EnsembleSelectorLayout(
                self.register_component_unique_id(self.Ids.ENSEMBLE),
                self.register_component_unique_id(self.Ids.REALIZATION),
                list(self._ensemble_paths.keys()),
            ),
            FilterSelectorLayout(self.register_component_unique_id(self.Ids.FORMATION)),
            MapSelectorLayout(
                self._color_scale_names,
                self.register_component_unique_id(self.Ids.PROPERTY),
                self.register_component_unique_id(self.Ids.STATISTIC),
                self.register_component_unique_id(self.Ids.COLOR_SCALE),
                self.register_component_unique_id(self.Ids.CM_MIN),
                self.register_component_unique_id(self.Ids.CM_MAX),
                self.register_component_unique_id(self.Ids.CM_MIN_AUTO),
                self.register_component_unique_id(self.Ids.CM_MAX_AUTO),
            ),
            ExperimentalFeaturesLayout(
                self.register_component_unique_id(self.Ids.PLUME_THRESHOLD),
                self.register_component_unique_id(self.Ids.PLUME_SMOOTHING),
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(self.Ids.REALIZATION).to_string(), "options"
            ),
            Output(self.component_unique_id(self.Ids.REALIZATION).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.ENSEMBLE).to_string(), "value"),
        )
        def set_realizations(ensemble: str) -> Tuple[List[Dict[str, Any]], List[int]]:
            rlz = [
                {"value": r, "label": str(r)}
                for r in self._ensemble_surface_providers[ensemble].realizations()
            ]
            return rlz, [rlz[0]["value"]]  # type: ignore

        @callback(
            Output(self.component_unique_id(self.Ids.FORMATION).to_string(), "options"),
            Output(self.component_unique_id(self.Ids.FORMATION).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.PROPERTY).to_string(), "value"),
            State(self.component_unique_id(self.Ids.ENSEMBLE).to_string(), "value"),
            State(self.component_unique_id(self.Ids.FORMATION).to_string(), "value"),
        )
        def set_formations(
            prop: str, ensemble: str, current_value: str
        ) -> Tuple[List[Dict[str, Any]], Optional[str]]:
            if ensemble is None:
                return [], None
            surface_provider = self._ensemble_surface_providers[ensemble]
            # Map
            prop_name = property_origin(MapAttribute(prop), self._map_attribute_names)
            surfaces = surface_provider.surface_names_for_attribute(prop_name)
            # Formation names
            formations = [{"label": v.title(), "value": v} for v in surfaces]
            picked_formation = None
            if len(formations) != 0:
                if current_value is None and self._initial_surface in surfaces:
                    picked_formation = self._initial_surface
                elif current_value in surfaces:
                    picked_formation = dash.no_update
                else:
                    picked_formation = formations[0]["value"]
            return formations, picked_formation

        @callback(
            Output(
                self.component_unique_id(self.Ids.STATISTIC).to_string(), "disabled"
            ),
            Input(self.component_unique_id(self.Ids.REALIZATION).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.PROPERTY).to_string(), "value"),
        )
        def toggle_statistics(realizations: List[int], attribute: str) -> bool:
            if len(realizations) <= 1:
                return True
            if MapAttribute(attribute) in (
                MapAttribute.SGAS_PLUME,
                MapAttribute.AMFG_PLUME,
            ):
                return True
            return False

        @callback(
            Output(self.component_unique_id(self.Ids.CM_MIN).to_string(), "disabled"),
            Output(self.component_unique_id(self.Ids.CM_MAX).to_string(), "disabled"),
            Input(self.component_unique_id(self.Ids.CM_MIN_AUTO).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.CM_MAX_AUTO).to_string(), "value"),
        )
        def set_color_range_data(
            min_auto: List[str], max_auto: List[str]
        ) -> Tuple[bool, bool]:
            return len(min_auto) == 1, len(max_auto) == 1


class FilterSelectorLayout(wcc.Selectors):
    def __init__(self, formation_id: str):
        super().__init__(
            label="Filter Settings",
            children=[
                "Formation",
                wcc.Dropdown(
                    id=formation_id,
                    clearable=False,
                ),
            ],
        )


class MapSelectorLayout(wcc.Selectors):
    _CM_RANGE = {
        "display": "flex",
        "flexDirection": "row",
    }

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        color_scale_names: List[str],
        property_id: str,
        statistic_id: str,
        colormap_id: str,
        cm_min_id: str,
        cm_max_id: str,
        cm_min_auto_id: str,
        cm_max_auto_id: str,
    ):
        super().__init__(
            label="Map Settings",
            open_details=True,
            children=[
                html.Div(
                    [
                        "Property",
                        wcc.Dropdown(
                            id=property_id,
                            options=_compile_property_options(),
                            value=MapAttribute.MIGRATION_TIME.value,
                            clearable=False,
                        ),
                        "Statistic",
                        wcc.Dropdown(
                            id=statistic_id,
                            options=[s.value for s in SurfaceStatistic],
                            value=SurfaceStatistic.MEAN,
                            clearable=False,
                        ),
                        html.Div(
                            style={
                                "height": "1px",
                                "backgroundColor": "lightgray",
                                "gridColumnStart": "span 2",
                            }
                        ),
                        "Color Scale",
                        dcc.Dropdown(
                            id=colormap_id,
                            options=color_scale_names,
                            value=color_scale_names[0],
                            clearable=False,
                        ),
                        "Minimum",
                        html.Div(
                            [
                                dcc.Input(id=cm_min_id, type="number"),
                                dcc.Checklist(
                                    ["Auto"],
                                    ["Auto"],
                                    id=cm_min_auto_id,
                                ),
                            ],
                            style=self._CM_RANGE,
                        ),
                        "Maximum",
                        html.Div(
                            [
                                dcc.Input(id=cm_max_id, type="number"),
                                dcc.Checklist(
                                    ["Auto"],
                                    ["Auto"],
                                    id=cm_max_auto_id,
                                ),
                            ],
                            style=self._CM_RANGE,
                        ),
                    ],
                )
            ],
        )


class ExperimentalFeaturesLayout(wcc.Selectors):
    def __init__(self, plume_threshold_id: str, plume_smoothing_id: str):
        super().__init__(
            label="Experimental",
            open_details=False,
            children=[
                html.Div(
                    children=[
                        html.Div("Plume Threshold"),
                        dcc.Input(
                            id=plume_threshold_id,
                            type="number",
                            min=0,
                            value=0.000001,
                            placeholder="Lower property threshold",
                            style={
                                "textAlign": "right",
                            },
                        ),
                    ],
                ),
                html.Div(
                    children=[
                        html.Div("Plume Smoothing"),
                        dcc.Input(
                            id=plume_smoothing_id,
                            type="number",
                            min=0,
                            value=0,
                            step=1,
                            placeholder="Smoothing [#pixels]",
                            style={
                                "textAlign": "right",
                            },
                        ),
                    ],
                ),
            ],
        )


class EnsembleSelectorLayout(wcc.Selectors):
    def __init__(self, ensemble_id: str, realization_id: str, ensembles: List[str]):
        super().__init__(
            label="Ensemble",
            open_details=True,
            children=[
                "Ensemble",
                wcc.Dropdown(
                    id=ensemble_id,
                    options=[dict(value=en, label=en) for en in ensembles],
                    value=ensembles[0],
                    clearable=False,
                ),
                "Realization",
                wcc.SelectWithLabel(
                    id=realization_id,
                    value=[],
                    multi=True,
                ),
            ],
        )


def _compile_property_options() -> List[Dict[str, Any]]:
    return [
        {"label": "SGAS", "value": "", "disabled": True},
        {
            "label": MapAttribute.MIGRATION_TIME.value,
            "value": MapAttribute.MIGRATION_TIME.value,
        },
        {"label": MapAttribute.MAX_SGAS.value, "value": MapAttribute.MAX_SGAS.value},
        {
            "label": MapAttribute.SGAS_PLUME.value,
            "value": MapAttribute.SGAS_PLUME.value,
        },
        {"label": "AMFG", "value": "", "disabled": True},
        {"label": MapAttribute.MAX_AMFG.value, "value": MapAttribute.MAX_AMFG.value},
        {
            "label": MapAttribute.AMFG_PLUME.value,
            "value": MapAttribute.AMFG_PLUME.value,
        },
    ]



from typing import List

import pandas as pd
import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class FilterOption(StrEnum):
    REMOVE_SENS_WITH_NO_IMPACT = "remove-sens-with-no-impact"


class Scale(StrEnum):
    REL_VALUE_PERC = "Relative value (%)"
    REL_VALUE = "Relative value"
    TRUE_VALUE = "True value"


class ViewSettings(SettingsGroupABC):
    class IDs(StrEnum):
        REFERENCE = "reference"
        SCALE = "scale"
        SENSITIVITIES = "sensitivities"
        FILTER_OPTIONS = "plot-options"

    def __init__(
        self,
        realizations: pd.DataFrame,
        reference: str = "rms_seed",
        allow_click: bool = False,
    ) -> None:
        super().__init__("View Settings")

        self.sensnames = list(realizations["SENSNAME"].unique())

        self.initial_reference = (
            reference if reference in self.sensnames else self.sensnames[0]
        )

        self.allow_click = allow_click

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(ViewSettings.IDs.REFERENCE),
                label="Reference",
                options=[{"label": r, "value": r} for r in self.sensnames],
                value=self.initial_reference,
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(ViewSettings.IDs.SCALE),
                label="Scale",
                options=[{"label": r, "value": r} for r in Scale],
                value=Scale.REL_VALUE_PERC,
                multi=False,
                clearable=False,
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(ViewSettings.IDs.SENSITIVITIES),
                label="Select sensitivities",
                options=[{"label": r, "value": r} for r in self.sensnames],
                value=self.sensnames,
                multi=True,
            ),
            wcc.Checklist(
                id=self.register_component_unique_id(ViewSettings.IDs.FILTER_OPTIONS),
                label="Filter options",
                options=[
                    {
                        "label": "Remove sensitivities with no impact",
                        "value": FilterOption.REMOVE_SENS_WITH_NO_IMPACT,
                    }
                ],
                value=[],
                labelStyle={"display": "block"},
            ),
        ]
